import hmac
import logging
import os
import shutil
from io import BytesIO

from embit import bip32, bip39, networks

from ..device import Device
from ..helpers import alias
from ..key import Key
from ..rpc import get_default_datadir
from ..specter_error import SpecterError
from ..util.base58 import decode_base58, encode_base58_checksum
from ..util.descriptor import AddChecksum
from ..util.xpub import convert_xpub_prefix, get_xpub_fingerprint
from ..util.mnemonic import mnemonic_to_root
from . import DeviceTypes

logger = logging.getLogger(__name__)


class BitcoinCore(Device):
    device_type = DeviceTypes.BITCOINCORE
    name = "Bitcoin Core (hot wallet)"
    icon = "bitcoincore_icon.svg"

    hot_wallet = True
    taproot_support = True

    # default sighash to use
    SIGHASH = "ALL"

    def __init__(self, *args, **kwargs):
        self._use_descriptors = None
        super().__init__(*args, **kwargs)

    def setup_device(self, file_password, wallet_manager):
        wallet_name = os.path.join(wallet_manager.rpc_path + "_hotstorage", self.alias)
        core_version = wallet_manager.rpc.getnetworkinfo().get("version", 0)
        use_descriptors = core_version >= 210000
        if use_descriptors:
            # it could fail if Core is compiled without sqlite
            try:
                wallet_manager.rpc.createwallet(
                    wallet_name, False, True, file_password or "", False, True
                )
                self._use_descriptors = True
            except:
                wallet_manager.rpc.createwallet(
                    wallet_name, False, True, file_password or "", False, True
                )
                self.taproot_support = False
        else:
            wallet_manager.rpc.createwallet(
                wallet_name, False, True, file_password or ""
            )
        rpc = wallet_manager.rpc.wallet(wallet_name)

    def use_descriptors(self, rpc):
        if self._use_descriptors is None:
            self._use_descriptors = rpc.getwalletinfo().get("descriptors", False)
            if not self._use_descriptors:
                self.taproot_support = False
        return self._use_descriptors

    def taproot_available(self, rpc):
        try:
            # currently only master branch supports tr() descriptors
            # TODO: replace to 220000
            core_version = rpc.getnetworkinfo().get("version", 0)
            info = rpc.getblockchaininfo()
            taproot_active = (core_version >= 219900) and (
                info.get("softforks", {}).get("taproot", {}).get("active", False)
            )
            taproot_support = self.use_descriptors(rpc) and taproot_active
            self.taproot_support = taproot_support
            return taproot_support
        except Exception as e:
            self.taproot_support = False
            logger.exception(e)
            return False

    def add_hot_wallet_keys(
        self,
        mnemonic,
        passphrase,
        paths,
        file_password,
        wallet_manager,
        testnet,
        keys_range=[0, 1000],
        keys_purposes=[],
    ):
        if type(keys_range[0]) == str:
            keys_range[0] = int(keys_range[0])
        if type(keys_range[1]) == str:
            keys_range[1] = int(keys_range[1])
        root = mnemonic_to_root(mnemonic, passphrase)
        network = networks.NETWORKS["test" if testnet else "main"]
        root.version = network["xprv"]
        xprv = root.to_base58()
        # Load the wallet if not loaded
        self._load_wallet(wallet_manager)
        rpc = wallet_manager.rpc.wallet(
            os.path.join(wallet_manager.rpc_path + "_hotstorage", self.alias)
        )
        if file_password:
            rpc.walletpassphrase(file_password, 60)

        # check if we can use descriptors and taproot
        use_descriptors = self.use_descriptors(rpc)
        taproot_available = self.taproot_available(rpc)

        args = [
            {
                "desc": AddChecksum(
                    "{}({}{}/0/*)".format(
                        "tr"
                        if path.startswith("m/86h") and taproot_available
                        else "wpkh",
                        xprv,
                        path.rstrip("/").replace("m", ""),
                    )
                ),
                "range": keys_range,
                "timestamp": "now",
            }
            for path in paths
        ] + [
            {
                "desc": AddChecksum(
                    "{}({}{}/1/*)".format(
                        "tr"
                        if path.startswith("m/86h") and taproot_available
                        else "wpkh",
                        xprv,
                        path.rstrip("/").replace("m", ""),
                    )
                ),
                "range": keys_range,
                "timestamp": "now",
            }
            for path in paths
        ]

        if use_descriptors:
            rpc.importdescriptors(args, timeout=15)
        else:
            rpc.importmulti(args, {"rescan": False})

        xpubs = [root.derive(path).to_public().to_base58() for path in paths]
        # root fingerprint is fingerprint field of the first child
        master_fpr = root.child(0).fingerprint.hex()
        keys = []
        for i in range(len(paths)):
            try:
                path = paths[i]
                xpub = xpubs[i]
                # detect slip132 version for xpubs
                slip132_prefix = bip32.detect_version(
                    path, default="xpub", network=network
                )
                xpub = "[{}{}]{}\n".format(
                    master_fpr,
                    path.replace("m", ""),
                    convert_xpub_prefix(xpub, slip132_prefix),
                )
                keys.append(
                    Key.parse_xpub(
                        xpub, keys_purposes[i] if len(keys_purposes) > i else ""
                    )
                )
            except Exception:
                # TODO: This should never occur, but just in case,
                # we must make sure to catch it properly so it
                # doesn't crash the app no matter what.
                raise Exception("Failed to parse this xpub:\n" + "\n".join(xpub))
        self.add_keys(keys)

    def _load_wallet(self, wallet_manager):
        try:
            existing_wallets = [
                w["name"] for w in wallet_manager.rpc.listwalletdir()["wallets"]
            ]
        except:
            existing_wallets = None
        loaded_wallets = wallet_manager.rpc.listwallets()

        hotstorage_path = wallet_manager.rpc_path + "_hotstorage"
        wallet_path = os.path.join(hotstorage_path, self.alias)
        if existing_wallets is None or wallet_path in existing_wallets:
            if wallet_path not in loaded_wallets:
                wallet_manager.rpc.loadwallet(wallet_path)

    def is_encrypted(self, wallet_manager):
        """Check if the wallet is encrypted"""
        self._load_wallet(wallet_manager)
        hotstorage_path = wallet_manager.rpc_path + "_hotstorage"
        wallet_path = os.path.join(hotstorage_path, self.alias)
        try:
            # check if password is enabled
            info = wallet_manager.rpc.getwalletinfo(wallet=wallet_path)
            return "unlocked_until" in info
        except Exception as e:
            logger.warning("Cannot fetch hot wallet info")
        # Assuming encrypted by default
        return True

    def create_psbts(self, base64_psbt, wallet):
        return {"core": base64_psbt}

    def sign_psbt(self, base64_psbt, wallet, file_password=None):
        # Load the wallet if not loaded
        self._load_wallet(wallet.manager)
        rpc = wallet.manager.rpc.wallet(
            os.path.join(wallet.manager.rpc_path + "_hotstorage", self.alias)
        )
        if file_password:
            rpc.walletpassphrase(file_password, 60)
        signed_psbt = rpc.walletprocesspsbt(base64_psbt, True, self.SIGHASH)
        if base64_psbt == signed_psbt["psbt"]:
            raise Exception(
                "Make sure you have entered the wallet file password correctly. (If your wallet is not encrypted submit empty password)"
            )
        if file_password:
            rpc.walletlock()
        return signed_psbt

    def sign_raw_tx(self, raw_tx, wallet, file_password=None):
        # Load the wallet if not loaded
        self._load_wallet(wallet.manager)
        rpc = wallet.manager.rpc.wallet(
            os.path.join(wallet.manager.rpc_path + "_hotstorage", self.alias)
        )
        if file_password:
            rpc.walletpassphrase(file_password, 60)
        signed_tx = rpc.signrawtransactionwithwallet(raw_tx)
        if file_password:
            rpc.walletlock()
        return signed_tx

    def delete(
        self, wallet_manager, bitcoin_datadir=get_default_datadir(), chain="main"
    ):
        try:
            wallet_rpc_path = os.path.join(
                wallet_manager.rpc_path + "_hotstorage", self.alias
            )
            wallet_manager.rpc.unloadwallet(wallet_rpc_path)
            # Try deleting wallet file
            if bitcoin_datadir:
                if chain != "main":
                    bitcoin_datadir = os.path.join(bitcoin_datadir, chain)
                candidates = [
                    os.path.join(bitcoin_datadir, wallet_rpc_path),
                    os.path.join(bitcoin_datadir, "wallets", wallet_rpc_path),
                ]
                for path in candidates:
                    if os.path.exists(path):
                        shutil.rmtree(path)
                        break
        except:
            pass  # We tried...


class BitcoinCoreWatchOnly(BitcoinCore):
    """If a BitcoinCore Hotwallet get exported, it'll have the type:"bitcoincore_watchonly" .
    if such a device.json get imported, it's instantiate as a BitcoinCoreWatchOnly.
    It can be converted back to a device of Type BitcoinCore by providing the 12 words again.
    """

    device_type = DeviceTypes.BITCOINCORE_WATCHONLY
    name = "Bitcoin Core (watch only)"
    hot_wallet = False

    def sign_psbt(self, base64_psbt, wallet, file_password=None):
        raise SpecterError("Cannot sign with a watch-only wallet. Convert")

    def sign_raw_tx(self, raw_tx, wallet, file_password=None):
        raise SpecterError("Cannot sign with a watch-only wallet")

    def add_hot_wallet_keys(
        self,
        mnemonic,
        passphrase,
        paths,
        file_password,
        wallet_manager,
        testnet,
        keys_range=[0, 1000],
        keys_purposes=[],
    ):
        # Convert the watch-only wallet to a hot wallet then fix up its internal attrs to
        # match.
        super().add_hot_wallet_keys(
            mnemonic,
            passphrase,
            paths,
            file_password,
            wallet_manager,
            testnet,
            keys_range,
            keys_purposes,
        )

        # Change type (also triggers write to file)
        self.set_type(DeviceTypes.BITCOINCORE)
        # After update this device will be available as a BitcoinCore (hot) instance
        self.manager.update()



from embit.liquid.pset import PSET, LInputScope, LOutputScope
from embit.liquid.transaction import LTransaction, LTransactionOutput
from embit.liquid.networks import get_network
from embit.liquid.addresses import address as liquid_address
from embit.liquid import slip77
from embit import bip32, ec, script
from math import ceil
import time
from cryptoadvance.specter.util.psbt import *


def to_canonical_pset(pset: str) -> str:
    """
    Removes unblinded information from the transaction
    so Elements Core can decode it
    """
    # if we got psbt, not pset - just return
    if not pset.startswith("cHNl"):
        return pset
    tx = PSET.from_string(pset)

    for inp in tx.inputs:
        inp.value = None
        inp.asset = None
        inp.value_blinding_factor = None
        inp.asset_blinding_factor = None

    for out in tx.outputs:
        if out.is_blinded:
            out.asset = None
            out.asset_blinding_factor = None
            out.value = None
            out.value_blinding_factor = None
    return str(tx)


def get_address(script_pubkey: script.Script, network: dict) -> str:
    if not script_pubkey.data:
        return "Fee"
    if script_pubkey.data.startswith(b"\x6a"):
        if len(script_pubkey.data) == 1:
            return "DUMMY"  # dummy output to blind
        else:
            return "OP_RETURN " + script_pubkey.data[1:].hex()
    try:
        return script_pubkey.address(network)
    except:
        return script_pubkey.data.hex()


def get_value(value) -> int:
    if isinstance(value, int):
        return value
    return 0  # confidential


def get_asset(asset) -> bytes:
    if len(asset) != 32:
        return (b"\xFF" * 32).hex()  # confidential
    return asset[::-1].hex()


class SpecterLTx(SpecterTx):
    TxCls = LTransaction

    def vout_to_dict(self, vout: LTransactionOutput) -> dict:
        i = self.tx.vout.index(vout)
        return {
            "value": round(1e-8 * get_value(vout.value), 8),
            "sats": get_value(vout.value),
            "n": i,
            "asset": get_asset(vout.asset),
            "scriptPubKey": {
                "hex": vout.script_pubkey.data.hex(),
                "addresses": [get_address(vout.script_pubkey, self.network)],
            },
        }


class SpecterLInputScope(SpecterInputScope):
    TxCls = SpecterLTx

    @property
    def assetid(self) -> str:
        if self.scope.asset is None:
            return "???"
        return self.scope.asset[::-1].hex()

    @property
    def address(self) -> str:
        # TODO: blinding key?
        try:
            return liquid_address(self.scope.script_pubkey, network=self.network)
        except:
            return None

    @property
    def sat_amount(self) -> int:
        return self.scope.value or 0

    def to_dict(self) -> dict:
        obj = super().to_dict()
        obj.update({"asset": self.assetid})
        return obj


class SpecterLOutputScope(SpecterOutputScope):
    @property
    def assetid(self) -> str:
        if self.scope.asset is None:
            return "???"
        return self.scope.asset[::-1].hex()

    @property
    def address(self) -> str:
        if not self.scope.script_pubkey.data:
            return "Fee"
        if self.scope.script_pubkey.data.startswith(b"\x6a"):
            if len(self.scope.script_pubkey.data) == 1:
                return "DUMMY"  # dummy output to blind
            else:
                return "OP_RETURN " + self.scope.script_pubkey.data[1:].hex()
        try:
            # try making a liquid address
            return liquid_address(
                self.scope.script_pubkey, self.blinding_key, network=self.network
            )
        except:
            # if failed - return hex of the scriptpubkey
            return self.scope.script_pubkey.data.hex()

    def extra_weight(self) -> int:
        wit = 0
        if self.scope.is_blinded:
            wit += 33 * 4  # nonce
            wit += (33 - 9) * 4  # value
            if self.scope.range_proof and self.scope.surjection_proof:
                # serialized witness length
                wit += (
                    len(self.scope.surjection_proof) + len(self.scope.range_proof) + 3
                )
            else:
                # we don't have proofs yet but we can estimate their size
                wit += 4245
        return wit

    @property
    def blinding_key(self) -> ec.PublicKey:
        if self.scope.blinding_pubkey:
            return ec.PublicKey.parse(self.scope.blinding_pubkey)

    @property
    def sat_amount(self) -> int:
        return self.scope.value or 0

    def to_dict(self) -> dict:
        obj = super().to_dict()
        obj.update({"asset": self.assetid})
        return obj


class SpecterPSET(SpecterPSBT):
    """Specter's PSBT class with some handy functions"""

    PSBTCls = PSET
    InputCls = SpecterLInputScope
    OutputCls = SpecterLOutputScope
    TxCls = SpecterLTx

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.psbt.verify(ignore_missing=True)

    @property
    def full_size(self) -> int:
        size = len(self.psbt.tx.serialize()) * 4
        # witness and redeem script
        size += len(self.inputs) * self.extra_input_weight
        for out in self.outputs:
            size += out.extra_weight()
        return ceil(size / 4)

    def should_display(self, out: SpecterLOutputScope) -> bool:
        """Checks if this output should be displayed"""
        if not out.scope.script_pubkey.data:
            # Fee
            return False
        if out.scope.value == 0 and out.scope.script_pubkey.data == b"\x6a":
            # Dummy output
            return False
        return super().should_display(out)

    @property
    def assets(self) -> List[str]:
        return [out.assetid for out in self.outputs if self.should_display(out)]

    def to_dict(self) -> dict:
        obj = super().to_dict()
        obj.update({"asset": self.assets})
        return obj



import base64
import datetime
import hashlib
import json
import logging
import secrets
from decimal import Decimal
from os import access
from typing import List
from urllib import request
from urllib.parse import urlparse

import pytz
import requests
from flask import current_app as app
from flask_babel import lazy_gettext as _

logger = logging.getLogger(__name__)


# TODO: Update with prod values

code_verifier = "64fRjTuy6SKqdC1wSoInUNxX65dQUhVVKTqZXuQ7dqw"


class SwanApiException(Exception):
    pass


class SwanApiRefreshTokenException(SwanApiException):
    pass


class SwanClient:
    def __init__(
        self, hostname, access_token: str, access_token_expires, refresh_token
    ):
        self.hostname = hostname
        self.access_token: str = str(
            access_token
        )  # The ServiceSwan storage might have one in its storage
        self.access_token_expires = (
            access_token_expires  # a timestamp when the access token will expire
        )
        self.refresh_token = refresh_token

        self.api_url = app.config.get("SWAN_API_URL")
        self.client_id = app.config.get("SWAN_CLIENT_ID")
        self.client_secret = app.config.get("SWAN_CLIENT_SECRET")

    def is_access_token_valid(self):
        return (
            self.access_token_expires > datetime.datetime.now(tz=pytz.utc).timestamp()
        )

    def calc_callback_url(self):
        return f"http://{ self.hostname }{app.config['APP_URL_PREFIX']}{app.config['EXT_URL_PREFIX']}/swan/oauth2/callback"

    def get_oauth2_start_url(self, callback_hostname):
        """
        Set up the Swan API integration by requesting our initial access_token and
        refresh_token.
        """
        # Let's start the PKCE-flow
        global code_verifier

        self.hostname = callback_hostname

        if code_verifier is None:
            code_verifier = secrets.token_urlsafe(43)
        # see specification: https://datatracker.ietf.org/doc/html/rfc7636#section-4.2
        # and example impl: https://github.com/RomeoDespres/pkce/blob/master/pkce/__init__.py#L94-L96
        hashed = hashlib.sha256(code_verifier.encode("ascii")).digest()
        encoded = base64.urlsafe_b64encode(hashed)
        code_challenge = encoded.decode("ascii")[:-1]

        flow_url = f"{self.api_url}/oidc/auth?"
        query_params = [
            f"client_id={self.client_id}",
            f"redirect_uri={self.calc_callback_url()}",
            "response_type=code",
            "response_mode=query",
            f"code_challenge={code_challenge}",
            "code_challenge_method=S256",
            "state=kjkmdskdmsmmsmdslmdlsm",
            "scope=offline_access v1 write:vendor_wallet read:vendor_wallet write:automatic_withdrawal read:automatic_withdrawal",
            "prompt=consent",
        ]
        flow_url += "&".join(query_params)

        return flow_url

    def _get_access_token(
        self, code: str = None, code_verifier: str = None, request_uri=None
    ) -> str:
        """
        If code and code_verifier are specified, this is our initial request for an
        access_token and, more importantly, the refresh_token.

        If code is None, use the refresh_token to get a new short-lived access_token.

        If we don't have the refresh_token, raise SwanApiRefreshTokenException.
        """
        # Must explicitly set User-Agent; Swan firewall blocks all requests with "python".
        auth_header = {"User-Agent": "Specter Desktop"}

        if code:
            # Requesting initial refresh_token and access_token
            payload = {
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "code_verifier": code_verifier,
                "grant_type": "authorization_code",
                "state": "kjkmdskdmsmmsmdslmdlsm",
                "code": code,
                "redirect_uri": request_uri,
            }
        else:
            if self.is_access_token_valid():
                return self.access_token

            # Use the refresh_token to get a new access_token
            if not self.refresh_token:
                raise SwanApiRefreshTokenException(
                    "access_token is expired but we don't have a refresh_token"
                )

            payload = {
                "grant_type": "refresh_token",
                # "redirect_uri":   # Necessary? Probably yes but more difficult to reconstruct it but the refresh-token is not yet used anyway
                "refresh_token": self.refresh_token,
                "scope": "offline_access v1 write:vendor_wallet read:vendor_wallet write:automatic_withdrawal read:automatic_withdrawal",
            }

            auth_hash = base64.b64encode(
                f"{self.client_id}:{self.client_secret}".encode()
            ).decode()
            auth_header["Authorization"] = f"Basic {auth_hash}"

        response = requests.post(
            f"{self.api_url}/oidc/token",
            data=payload,
            headers=auth_header,
        )
        resp = json.loads(response.text)
        """
            {
                "access_token": "***************",
                "expires_in": 3600,
                "refresh_token": "***************",
                "scope": "offline_access v1 write:vendor_wallet read:vendor_wallet write:automatic_withdrawal read:automatic_withdrawal",
                "token_type": "Bearer"
            }
        """
        if resp.get("access_token"):
            self.access_token = resp.get("access_token")
            self.access_token_expires = (
                datetime.datetime.now(tz=pytz.utc)
                + datetime.timedelta(seconds=resp["expires_in"])
            ).timestamp()
            self.refresh_token = resp.get("refresh_token")
            return self.access_token
        else:
            logger.warning(response)
            raise SwanApiException(response.text)

    def handle_oauth2_auth_callback(self, request):
        code = request.args.get("code")
        rp = urlparse(request.url)
        request_uri = f"{rp.scheme}://{rp.netloc}{rp.path}"
        self._get_access_token(
            code=code, code_verifier=code_verifier, request_uri=request_uri
        )

    def authenticated_request(
        self, endpoint: str, method: str = "GET", json_payload: dict = {}
    ) -> dict:
        logger.debug(f"{method} endpoint: {endpoint}")

        access_token = self._get_access_token()

        # Must explicitly set User-Agent; Swan firewall blocks all requests with "python".
        auth_header = {
            "User-Agent": "Specter Desktop",
            "Authorization": f"Bearer {access_token}",
        }
        try:
            if method == "GET":
                response = requests.get(self.api_url + endpoint, headers=auth_header)
            elif method in ["POST", "PATCH", "PUT", "DELETE"]:
                response = requests.request(
                    method=method,
                    url=self.api_url + endpoint,
                    headers=auth_header,
                    json=json_payload,
                )
            if response.status_code != 200:
                raise SwanApiException(f"{response.status_code}: {response.text}")
            return response.json()
        except Exception as e:
            # TODO: tighten up expected Exceptions
            logger.exception(e)
            logger.error(
                f"endpoint: {self.api_url}{endpoint} | method: {method} | payload: {json.dumps(json_payload, indent=4)}"
            )
            logger.error(f"{response.status_code}: {response.text}")
            raise e

    def get_autowithdrawal_addresses(self, swan_wallet_id: str) -> dict:
        """
        {
            "entity": "wallet",
            "item": {
                "id": "c47e1e83-90a0-45da-ae25-6a0d324b9f29",
                "isConfirmed": false,
                "displayName": "Specter autowithdrawal to SeedSigner demo",
                "metadata": {
                    "oidc": {
                        "clientId": "specter-dev"
                    },
                    "specter_wallet_alias": "seedsigner_demo"
                },
                "btcAddresses": []
            }
        }
        """

        resp = self.authenticated_request(
            endpoint=f"/apps/v20210824/wallets/{swan_wallet_id}?full=true",
            method="GET",
        )

        return resp

    def update_autowithdrawal_addresses(
        self,
        swan_wallet_id: str,
        specter_wallet_name: str,
        specter_wallet_alias: str,
        addresses: List[str],
    ) -> dict:
        """
        * If SWAN_WALLET_ID is known, any existing unused addresses are cleared.
        * If there is no known SWAN_WALLET_ID, we `POST` to create an initial Swan wallet and store the resulting SWAN_WALLET_ID.
        * Sends the list of new addresses for SWAN_WALLET_ID.
        """

        # normalize the strucure compatible with what swan will accept:
        # like: [{"address": "bcrt1q8k8a73crvjs06jhdj7xee8mace3mhlxj4pdvna"}, {"address": "bcrt ...
        addresses = [address["address"] for address in addresses]

        if swan_wallet_id:
            # We already have a Swan walletId. DELETE the existing unused addresses...
            self.delete_autowithdrawal_addresses(swan_wallet_id)

            # ...and then append the new ones.
            endpoint = f"/apps/v20210824/wallets/{swan_wallet_id}/addresses"
            method = "PATCH"
        else:
            # We don't yet have a Swan walletId. POST to create one.
            endpoint = "/apps/v20210824/wallets"
            method = "POST"

        resp = self.authenticated_request(
            endpoint=endpoint,
            method=method,
            json_payload={
                "btcAddresses": [{"address": addr} for addr in addresses],
                "displayName": str(
                    _('Specter Desktop "{}"').format(specter_wallet_name)
                ),  # Can't pass a LazyString into json
                "metadata": {
                    "specter_wallet_alias": specter_wallet_alias,
                },
            },
        )

        """
        Response should include wallet ("item") details:
            {
                "entity": "wallet",
                "item": {
                    "id": "c47e1e83-90a0-45da-ae25-6a0d324b9f29",
                    "isConfirmed": false,
                    "displayName": "Specter autowithdrawal to SeedSigner demo",
                    "metadata": {
                        "oidc": {
                            "clientId": "specter-dev"
                        },
                        "specter_wallet_alias": "seedsigner_demo"
                    }
                }
            }
        """
        if "item" in resp and "id" in resp["item"]:
            if resp["item"]["id"] != swan_wallet_id:
                swan_wallet_id = resp["item"]["id"]
                return swan_wallet_id
        else:
            raise SwanApiException(
                f"No 'id' returned for the new/updated wallet: {json.dumps(resp, indent=4)}"
            )

    def delete_autowithdrawal_addresses(self, swan_wallet_id: str):
        """
        Deletes all unused autowithdrawal addresses from the specified SWAN_WALLET_ID
        """
        resp = self.authenticated_request(
            endpoint=f"/apps/v20210824/wallets/{swan_wallet_id}/addresses",
            method="DELETE",
        )
        return resp

    def get_autowithdrawal_info(self) -> dict:
        """
        See note in set_autowithdrawal. This returns all autowithdrawal objs from the Swan
        side.
        """
        resp = self.authenticated_request(
            endpoint="/apps/v20210824/automatic-withdrawal",
            method="GET",
        )
        return resp

    def set_autowithdrawal(self, swan_wallet_id, btc_threshold: Decimal) -> dict:
        """
        0 == Weekly; other float values = BTC threshold

        The Swan api generates a new autowithdrawal id each time but there is no support to
        update an existing autowithdrawal, other than activating or deactivating it.

        New autowithdrawals are initialized as `isActive: false` and require the user to
        complete a Swan-side email verification step.

        We save the resulting autowithdrawal_id even though it isn't clear at the moment if
        it's desirable to ever call the `deactivate/` or `activate/` endpoints.
        """

        endpoint = "/apps/v20210824/automatic-withdrawal"
        method = "POST"
        resp = self.authenticated_request(
            endpoint=endpoint,
            method=method,
            json_payload={
                "walletId": swan_wallet_id,
                "minBtcThreshold": btc_threshold,
            },
        )

        """
            {
                "entity": "automaticWithdrawal",
                "item": {
                    "id": "******************",
                    "minBtcThreshold": "0.01",
                    "isActive": false,
                    "isCanceled": false,
                    "createdAt": "2022-01-07T02:14:56.070Z",
                    "walletId": "******************",
                    "walletAddressId": null
                }
            }
        """
        if "item" in resp and "id" in resp["item"]:
            return resp
        else:
            raise SwanApiException(
                f"No 'id' returned for the new/updated autowithdrawal: {json.dumps(resp, indent=4)}"
            )

    def activate_autowithdrawal(self, autowithdrawal_id) -> dict:
        """
        Activates the autowithdrawal specified in SwanService.AUTOWITHDRAWAL_ID.

        If the automatic withdrawal was just created, this will generate a 400 error:
        "Cannot activate an automatic withdrawal before withdrawal address is confirmed".

        The user must first confirm the first withdrawal addr via Swan-side email flow.
        After they confirm, the autowithdrawal should then return `isActive: true`.

        NOT CURRENTLY USED; remove if we don't ever enable disable/activate flows.
        """

        endpoint = f"/apps/v20210824/automatic-withdrawal/{autowithdrawal_id}/activate"
        method = "POST"
        resp = self.authenticated_request(
            endpoint=endpoint,
            method=method,
        )

        """
            {
                "entity": "automaticWithdrawal",
                "item": {
                    "id": "******************",
                    "minBtcThreshold": "0.01",
                    "isActive": true,
                    "isCanceled": false,
                    "createdAt": "2022-01-07T02:14:56.070Z",
                    "walletId": "******************",
                    "walletAddressId": null
                }
            }
        """
        if "item" in resp and "id" in resp["item"]:
            autowithdrawal_id = resp["item"]["id"]
            return autowithdrawal_id
        raise SwanApiException(
            f"No 'id' returned for the new/updated autowithdrawal: {json.dumps(resp, indent=4)}"
        )

    def get_wallet_details(self, swan_wallet_id: str) -> dict:
        """
        {
            "entity": "wallet",
            "item": {
                "id": "********************",
                "walletAddressId": ""********************",
                "btcAddress": ""********************",
                "isConfirmed": true,
                "displayName": "Specter Desktop \"DCA Cold Storage\"",
                "metadata": {
                    "oidc": {
                        "clientId": "specter-dev"
                    },
                    "specter_wallet_alias": "dca_cold_storage"
                }
            }
        }
        """
        resp = self.authenticated_request(
            endpoint=f"/apps/v20210824/wallets/{swan_wallet_id}",
            method="GET",
        )

        return resp

    def get_wallets(self) -> dict:
        """
        Return all Swan wallet entries. Should only be one per Specter-Swan user combo (but can be more due
        to testing/debugging, calling `/wallets` POST more than once, etc.)
        """
        resp = self.authenticated_request(
            endpoint=f"/apps/v20210824/wallets",
            method="GET",
        )
        """
            {
                "entity": "wallet",
                "list": [
                    {
                        "id": "**********",
                        "walletAddressId": "**********",
                        "btcAddress": "bc1q**********",
                        "isConfirmed": false,
                        "displayName": "Specter Desktop \"SeedSigner demo\"",
                        "metadata": {
                            "oidc": {
                                "clientId": "specter-dev"
                            },
                            "specter_wallet_alias": "seedsigner_demo"
                        }
                    },
                    {
                        "id": "**********",
                        "walletAddressId": "**********",
                        "btcAddress": "bc1q**********",
                        "isConfirmed": false,
                        "displayName": "Specter Desktop \"DCA Corn\"",
                        "metadata": {
                            "oidc": {
                                "clientId": "specter-dev"
                            },
                            "specter_wallet_alias": "dca_corn_2"
                        }
                    },
                    ...,
                ]
            }
        """
        return resp



import logging
from flask import Flask, Response, redirect, render_template, request, url_for, flash
from flask_login import login_required

from ..controller import user_secret_decrypted_required
from .service import BitcoinReserveService


"""
    Empty placeholder just so the dummyservice/static folder can be wired up to retrieve its img
"""

logger = logging.getLogger(__name__)

bitcoinreserve_endpoint = BitcoinReserveService.blueprint


@bitcoinreserve_endpoint.route("/")
@login_required
@user_secret_decrypted_required
def index():
    return render_template(
        "bitcoinreserve/index.jinja",
    )



import hashlib
from . import DeviceTypes
from .sd_card_device import SDCardDevice
from .hwi.specter_diy import enumerate as specter_enumerate, SpecterClient
from ..helpers import to_ascii20
from embit import bip32
from embit.psbt import PSBT, DerivationPath
from embit.liquid.pset import PSET
from embit.liquid.transaction import LSIGHASH
from binascii import a2b_base64, b2a_base64
from embit.liquid.networks import get_network
from embit.hashes import hash160
import logging

logger = logging.getLogger(__name__)


def fill_external_wallet_derivations(psbt, wallet):
    # fill derivations for receiving wallets if they own address
    net = get_network(wallet.manager.chain)
    wallets = wallet.manager.wallets.values()
    for out in psbt.outputs:
        try:  # if we fail - not a big deal
            # check if output derivation is empty
            if out.bip32_derivations:
                continue
            try:
                # if there is no address representation - continue
                addr = out.script_pubkey.address(net)
            except:
                continue
            for w in wallets:
                # skip sending wallet
                if w == wallet:
                    continue
                if not w.is_address_mine(addr):
                    continue
                # we get here if wallet owns address
                info = w.get_address_info(addr)
                derivation = [int(info.change), int(info.index)]
                # we first one key and build derivation for it
                # it's enough for DIY to do the res
                k = w.keys[0]
                key = bip32.HDKey.from_string(k.xpub)
                if k.fingerprint != "":
                    fingerprint = bytes.fromhex(k.fingerprint)
                else:
                    fingerprint = key.my_fingerprint
                if k.derivation != "":
                    der = bip32.parse_path(k.derivation)
                else:
                    der = []
                pub = key.derive(derivation).key
                out.bip32_derivations[pub] = DerivationPath(
                    fingerprint, der + derivation
                )
                break
        except Exception as e:
            logger.exception(e)


class Specter(SDCardDevice):
    device_type = DeviceTypes.SPECTERDIY
    name = "Specter-DIY"
    icon = "specter_icon.svg"

    exportable_to_wallet = True
    sd_card_support = True
    qr_code_support = True
    qr_code_support_verify = True
    wallet_export_type = "qr"
    supports_hwi_multisig_display_address = True
    liquid_support = True
    taproot_support = True

    def create_psbts(self, base64_psbt, wallet):
        # liquid transaction
        if base64_psbt.startswith("cHNl"):
            # remove rangeproofs and add sighash alls
            psbt = PSET.from_string(base64_psbt)
            # make sure we have tx blinding seed in the transaction
            if psbt.unknown.get(b"\xfc\x07specter\x00"):
                for out in psbt.outputs:
                    out.range_proof = None
                    # out.surjection_proof = None
                    # we know assets - we can blind it
                    if out.asset:
                        out.asset_commitment = None
                        out.asset_blinding_factor = None
                    # we know value - we can blind it
                    if out.value:
                        out.value_commitment = None
                        out.value_blinding_factor = None
                for inp in psbt.inputs:
                    if inp.value and inp.asset:
                        inp.range_proof = None
        else:
            psbt = PSBT.from_string(base64_psbt)

        fill_external_wallet_derivations(psbt, wallet)

        base64_psbt = psbt.to_string()
        psbts = super().create_psbts(base64_psbt, wallet)
        # remove non-witness utxo if they are there to reduce QR code size
        updated_psbt = wallet.fill_psbt(
            base64_psbt, non_witness=False, xpubs=False, taproot_derivations=True
        )
        try:
            qr_psbt = PSBT.from_string(updated_psbt)
        except:
            qr_psbt = PSET.from_string(updated_psbt)
        # find my key
        fgp = None
        derivation = None
        for k in wallet.keys:
            if k in self.keys and k.fingerprint and k.derivation:
                fgp = bytes.fromhex(k.fingerprint)
                derivation = bip32.parse_path(k.derivation)
                break
        # remove unnecessary derivations from inputs and outputs
        for inp in qr_psbt.inputs + qr_psbt.outputs:
            # keep only one derivation path (idealy ours)
            found = False
            pubkeys = list(inp.bip32_derivations.keys())
            for i, pub in enumerate(pubkeys):
                if fgp and inp.bip32_derivations[pub].fingerprint != fgp:
                    # only pop if we already saw our derivation
                    # or if it's not the last one
                    if found or i < len(pubkeys) - 1:
                        inp.bip32_derivations.pop(k, None)
                else:
                    found = True
        # remove scripts from outputs (DIY should know about the wallet)
        for out in qr_psbt.outputs:
            out.witness_script = None
            out.redeem_script = None
        # remove partial sigs from inputs
        for inp in qr_psbt.inputs:
            inp.partial_sigs = {}
        psbts["qrcode"] = qr_psbt.to_string()

        # we can add xpubs to SD card, but non_witness can be too large for MCU
        psbts["sdcard"] = wallet.fill_psbt(
            base64_psbt, non_witness=False, xpubs=True, taproot_derivations=True
        )
        psbts["hwi"] = wallet.fill_psbt(
            base64_psbt, non_witness=False, xpubs=True, taproot_derivations=True
        )
        return psbts

    def export_wallet(self, wallet):
        return (
            "addwallet "
            + to_ascii20(wallet.name)
            + "&"
            + get_wallet_qr_descriptor(wallet)
        )

    @classmethod
    def enumerate(cls, *args, **kwargs):
        return specter_enumerate(*args, **kwargs)

    @classmethod
    def get_client(cls, *args, **kwargs):
        return SpecterClient(*args, **kwargs)


def get_wallet_qr_descriptor(wallet):
    return wallet.recv_descriptor.split("#")[0].replace("/0/*", "")


def get_wallet_fingerprint(wallet):
    """
    Unique fingerprint of the wallet -
    first 4 bytes of hash160 of its descriptor
    """
    h160 = hash160(get_wallet_qr_descriptor(wallet).encode())
    return h160[:4]



# source: http://stackoverflow.com/questions/2758159/how-to-embed-a-python-interpreter-in-a-pyqt-widget

import traceback
import logging

logger = logging.getLogger(__name__)


class Console:
    def __init__(self):
        self.namespace = {}
        self.updateNamespace({"run": self.run_script})

    def run_script(self, filename):
        with open(filename) as f:
            script = f.read()
        return self.exec_command(script)

    def updateNamespace(self, namespace):
        self.namespace.update(namespace)
        # adding help message with the available keys
        self.namespace.update(
            {
                "help": f"You can try calling any of the following objects: {list(self.namespace.keys())}\n"
                f"Use a trailing '.' to get available methods and propeties, e.g. try: '{list(self.namespace.keys())[-1]}.'\n"
                "The command 'run' can execute python scripts, e.g., run('myscript.py')"
            }
        )

    def exec_command(self, command):
        if callable(self.namespace.get(command)):
            return "'{}' is a function. Type '{}()' to use it in the Python console.".format(
                command, command
            )

        try:
            try:
                # eval is generally considered bad practice. use it wisely!
                logger.info(f'Executing console command "{command}"')
                if command.endswith("."):
                    return {
                        "vars": eval(
                            f"vars({command[:-1]})", self.namespace, self.namespace
                        ),
                        "dir": eval(
                            f"dir({command[:-1]})", self.namespace, self.namespace
                        ),
                    }
                return eval(command, self.namespace, self.namespace)
            except SyntaxError:
                # exec is generally considered bad practice. use it wisely!
                return exec(command, self.namespace, self.namespace)
        except BaseException:
            traceback_lines = traceback.format_exc().split("\n")
            # Remove traceback mentioning this file, and a linebreak
            for i in (3, 2, 1, -1):
                traceback_lines.pop(i)
            return "\n".join(traceback_lines)



SETUP_STATES = {
    "STARTING_SETUP": {
        "bitcoind": "Starting Bitcoin Core Setup...",
        "torbrowser": "Starting Tor setup process...",
    },
    "VERIFY_SIGS": {
        "bitcoind": "Verifying signatures...",
        "torbrowser": "Verifying signatures...",
    },
    "START_SERVICE": {
        "bitcoind": "Starting up Bitcoin Core...",
        "torbrowser": "Starting up Tor daemon...",
    },
    "FAILED": {
        "bitcoind": "Failed to start Bitcoin Core...",
        "torbrowser": "Failed to start Tor daemon...",
    },
}



"""
Swan API uses PKCE OAuth2. Per Swan's API team: The client secret here is not considered
to be a real secret. There is no reasonable attack vector for this secret being public.
"""


class BaseConfig:
    SWAN_CLIENT_ID = "specter-dev"
    SWAN_CLIENT_SECRET = (
        "BcetcVcmueWf5P3UPJnHhCBMQ49p38fhzYwM7t3DJGzsXSjm89dDR5URE46SY69j"
    )
    SWAN_API_URL = "https://dev-api.swanbitcoin.com"
    SWAN_FRONTEND_URL = "https://dev-app.swanbitcoin.com/signup"

    # There is a whitelist at the oauth2-infra of swan which is hopefully in sync with this list:
    SWAN_ALLOWED_SPECTER_HOSTNAMES = [
        "localhost:25441",
        "umbrel.local:25441",
        "citadel.local:25441",
        "specter.local:25441",
    ]


class ProductionConfig(BaseConfig):
    SWAN_CLIENT_ID = "specter"
    SWAN_CLIENT_SECRET = (
        "UcqMZw3D70#E*Zo1hnC8f8P^Ils^6wligXMB*vL1fX@DYm6zloDI#p9Eemk8!y9#"
    )
    SWAN_API_URL = "https://api.swanbitcoin.com"
    SWAN_FRONTEND_URL = "https://www.swanbitcoin.com/Specter/"



import json
import logging
import requests
from cryptoadvance.specter.managers.wallet_manager import WalletManager

from cryptoadvance.specter.specter_error import SpecterError

from embit.descriptor import Descriptor
from embit.descriptor import Key as DescriptorKey
from embit.descriptor.arguments import AllowedDerivation
from embit.liquid.descriptor import LDescriptor
from cryptoadvance.specter.key import Key
from flask import flash
from flask_babel import lazy_gettext as _

logger = logging.getLogger(__name__)


class WalletImporter:
    """A class to create Wallets easily by json"""

    def __init__(self, wallet_json, specter, device_manager=None):
        """this will analyze the wallet_json and specifies self. ...:
        * wallet_name
        * recv_descriptor
        * cosigners_types
        from recv_descriptor and specter.chain:
        * descriptor
        from descriptor:
        * keys
        * cosigners
        * unknown_cosigners
        * unknown_cosigners_types
        """
        DescriptorCls = LDescriptor if specter.is_liquid else Descriptor
        if device_manager is None:
            device_manager = specter.device_manager
        try:
            self.wallet_data = json.loads(wallet_json)
            (
                self.wallet_name,
                recv_descriptor,
                self.cosigners_types,
            ) = WalletImporter.parse_wallet_data_import(self.wallet_data)
        except Exception as e:
            logger.warning(f"Trying to import: {wallet_json}")
            raise SpecterError(f"Unsupported wallet import format:{e}")
        try:
            self.descriptor = DescriptorCls.from_string(recv_descriptor)
            self.check_descriptor()
        except Exception as e:
            raise SpecterError(f"Invalid wallet descriptor: {e}")
        if self.wallet_name in specter.wallet_manager.wallets_names:
            raise SpecterError(f"Wallet with the same name already exists")
        (
            self.keys,
            self.cosigners,
            self.unknown_cosigners,
            self.unknown_cosigners_types,
        ) = self.parse_signers(device_manager.devices, self.cosigners_types)
        self.wallet_type = "multisig" if self.descriptor.is_basic_multisig else "simple"

    def check_descriptor(self):
        # Sparrow fix: if all keys have None as allowed derivation - set allowed derivation to [0, None]
        if all(
            [
                k.allowed_derivation is None or k.allowed_derivation.indexes == []
                for k in self.descriptor.keys
                if k.is_extended
            ]
        ):
            for k in self.descriptor.keys:
                if k.is_extended:
                    k.allowed_derivation = AllowedDerivation([0, None])

        # Check that all keys are HD keys and all have default derivation
        for key in self.descriptor.keys:
            if not key.is_extended:
                raise SpecterError("Only HD keys are supported in descriptor")
            if key.allowed_derivation is None or key.allowed_derivation.indexes != [
                0,
                None,
            ]:
                raise SpecterError(
                    "Descriptor key has wrong derivation, only /0/* derivation is supported."
                )

    def parse_signers(self, devices, cosigners_types):
        keys = []
        cosigners = []
        unknown_cosigners = []
        unknown_cosigners_types = []

        for i, descriptor_key in enumerate(self.descriptor.keys):
            # remove derivation from the key for comparison
            account_key = DescriptorKey.from_string(str(descriptor_key))
            account_key.allowed_derivation = None
            # Specter Key class
            desc_key = Key.parse_xpub(str(account_key))
            cosigner_found = False
            for cosigner in devices.values():
                for key in cosigner.keys:
                    # check key matches
                    if key.to_string(slip132=False) == desc_key.to_string(
                        slip132=False
                    ):
                        keys.append(key)
                        cosigners.append(cosigner)
                        cosigner_found = True
                        break
                if cosigner_found:
                    break
            if not cosigner_found:
                if len(cosigners_types) > i:
                    unknown_cosigners.append((desc_key, cosigners_types[i]["label"]))
                else:
                    unknown_cosigners.append((desc_key, None))
                if len(unknown_cosigners) > len(cosigners_types):
                    unknown_cosigners_types.append("other")
                else:
                    unknown_cosigners_types.append(cosigners_types[i]["type"])

        return (keys, cosigners, unknown_cosigners, unknown_cosigners_types)

    def create_nonexisting_signers(self, device_manager, request_form):
        """creates non existinging signer via the device_manager
        request_form is a dict which needs (where X is an integer):
        * unknown_cosigner_X_name
        * unknown_cosigner_X_type
        Apart from adding the devices, it'll also modify:
        * self.keys and
        * self.cosigners
        """
        for i, (unknown_cosigner_key, label) in enumerate(self.unknown_cosigners):
            unknown_cosigner_name = request_form["unknown_cosigner_{}_name".format(i)]
            unknown_cosigner_type = request_form.get(
                "unknown_cosigner_{}_type".format(i), "other"
            )

            device = device_manager.add_device(
                name=unknown_cosigner_name,
                device_type=unknown_cosigner_type,
                keys=[unknown_cosigner_key],
            )
            logger.info(f"Creating device {device}")
            self.keys.append(unknown_cosigner_key)
            self.cosigners.append(device)

    @property
    def address_type(self):
        if self.descriptor.is_taproot:
            return "tr"
        res = ""
        if self.descriptor.miniscript:
            if self.descriptor.wsh:
                res = "wsh"
        else:
            if self.descriptor.wpkh:
                res = "wpkh"
            else:
                return "pkh"
        if self.descriptor.sh:
            if res:
                return f"sh-{res}"
            else:
                return "sh"
        return res

    @property
    def sigs_required(self):
        sigs_required = 1
        if self.descriptor.is_basic_multisig:
            sigs_required = self.descriptor.miniscript.args[0].num
        return sigs_required

    @property
    def sigs_total(self):
        return len(self.descriptor.keys)

    def create_wallet(self, wallet_manager):
        """creates the wallet. Assumes all devices are there (create with create_nonexisting_signers)
        will also keypoolrefill and import_labels
        """
        try:
            kwargs = {}
            if (
                isinstance(self.descriptor, LDescriptor)
                and self.descriptor.blinding_key
            ):
                kwargs["blinding_key"] = self.descriptor.blinding_key.key

            self.wallet = wallet_manager.create_wallet(
                name=self.wallet_name,
                sigs_required=self.sigs_required,
                key_type=self.address_type,
                keys=self.keys,
                devices=self.cosigners,
                **kwargs,
            )
        except Exception as e:
            raise SpecterError(f"Failed to create wallet: {e}")
        logger.info(f"Created Wallet {self.wallet}")
        self.wallet.keypoolrefill(0, self.wallet.IMPORT_KEYPOOL, change=False)
        self.wallet.keypoolrefill(0, self.wallet.IMPORT_KEYPOOL, change=True)
        self.wallet.import_labels(self.wallet_data.get("labels", {}))
        return self.wallet

    def rescan_as_needed(self, specter):
        """will rescan the created wallet"""
        if not hasattr(self, "wallet"):
            raise Exception("called rescan_as_needed before create_wallet")
        potential_errors = None
        try:
            # get min of the two
            # if the node is still syncing
            # and the first block with tx is not there yet
            startblock = min(
                self.wallet_data.get("blockheight", specter.info.get("blocks", 0)),
                specter.info.get("blocks", 0),
            )
            # check if pruned
            if specter.info.get("pruned", False):
                newstartblock = max(startblock, specter.info.get("pruneheight", 0))
                if newstartblock > startblock:
                    potential_errors = SpecterError(
                        f"Using pruned node - we will only rescan from block {newstartblock}"
                    )
                    startblock = newstartblock
            self.wallet.rpc.rescanblockchain(startblock, no_wait=True)
            logger.info("Rescanning Blockchain ...")
        except Exception as e:
            logger.error("Exception while rescanning blockchain: %r" % e)
            if potential_errors:
                potential_errors = SpecterError(
                    str(potential_errors)
                    + " and "
                    + "Failed to perform rescan for wallet: %r" % e
                )
        self.wallet.getdata()
        if potential_errors:
            raise potential_errors

    @property
    def wallet_json(self):
        """self.wallet_data in json format"""
        return json.dumps(self.wallet_data)

    @classmethod
    def parse_wallet_data_import(cls, wallet_data):
        """Parses wallet JSON for import, takes JSON in a supported format
        and returns a tuple of wallet name, wallet descriptor, and cosigners types (electrum
        and newer Specter backups).
        Supported formats: Specter, Electrum, Account Map (Fully Noded, Gordian, Sparrow etc.)
        """
        cosigners_types = []

        # Specter-Desktop format
        if "recv_descriptor" in wallet_data:
            wallet_name = wallet_data.get("name", "Imported Wallet")
            recv_descriptor = wallet_data.get("recv_descriptor", None)

        # Electrum multisig
        elif "x1/" in wallet_data:
            i = 1
            xpubs = ""
            while "x{}/".format(i) in wallet_data:
                d = wallet_data["x{}/".format(i)]
                xpubs += "[{}]{}/0/*,".format(
                    d["derivation"].replace("m", d["root_fingerprint"]), d["xpub"]
                )
                if "hw_type" in d:
                    cosigners_types.append({"type": d["hw_type"], "label": d["label"]})
                else:  # this can occcur if no hardware wallet was used, but the seed is available
                    cosigners_types.append(
                        {"type": "electrum", "label": f"Electrum Multisig {i}"}
                    )
                    if "seed" in d:
                        flash(
                            _(
                                "The Electrum wallet contains a seed. The seed will not be imported."
                            ),
                            "warning",
                        )
                i += 1
            xpubs = xpubs.rstrip(",")

            if "xpub" in wallet_data["x1/"]:
                wallet_type = WalletImporter.wallet_type_by_slip132_xpub(
                    wallet_data["x1/"]["xpub"]
                )
            else:
                raise Exception('"xpub" not found in "x1/" in Electrum backup json')

            required_sigs = int(wallet_data.get("wallet_type").split("of")[0])
            recv_descriptor = "{}(sortedmulti({},{}))".format(
                wallet_type, required_sigs, xpubs
            )
            wallet_name = "Electrum {} of {}".format(required_sigs, i - 1)

        # Electrum singlesig
        # Here's a trimmed version of what an example Electrum file looks like (varies for different wallet types)
        # {
        #     "addr_history": {
        #         "tb1q2zqu0rxg70scs9umwwkjxvq7zdcle2ny6pda8v": [],
        #         "tb1q3p86mshyyrmzp2mw50eqeefxtkz8qfl8tsfgrs": [],
        #         "tb1q46u2rm20rp9rzauxxhdefregggag8cpelh0sqa": [
        #             [
        #                 "14f6bd3dec07c15773b29b4c7f7d3c913197840dfdf567b6b078e224f021b4ac",
        #                 2035683
        #             ],
        #             [
        #                 "23196c7657b60d1fde716c789881ffeee93554b9040c200589c65ceeac051cfb",
        #                 2035683
        #             ]
        #         ],
        #         "tb1q4k59fgwhwey8k4rs2l80p73dwtza0z0gvpjjfc": [],
        #         "tb1q56f5wgc7wqeqygpfgua64gucafp3wp6293ggcg": [],
        #         "tb1q8uysvaz8dysekwws868knvxngh5qfjzrlxqf2j": [
        #             [
        #                 "23196c7657b60d1fde716c789881ffeee93554b9040c200589c65ceeac051cfb",
        #                 2035683
        #             ]
        #         ],
        #         "tb1q8ydkp6qhq79ktwpcwjgt9dl7eer32vpw8gvduy": [
        #             [
        #                 "c9a54c381293a387a8c1c90e542d11cccf42cd0aeab20be3d1fcd10b465dcfde",
        #                 0
        #             ]
        #         ],
        #         "tb1qazyz882jeruqlppfralemm3r07pw6cpk426fhw": [],
        #         "tb1qch2j8ks2qt6thtf7wdpk6v9gquu8tfca6lsh0n": [],
        #         "tb1qcx2dultr5x9dc6jz0ye3wq240xfsdxnnl7vpsm": [
        #             [
        #                 "c9a54c381293a387a8c1c90e542d11cccf42cd0aeab20be3d1fcd10b465dcfde",
        #                 0
        #             ]
        #         ],
        #         "tb1qe686y2y39ncnafpzxxuyfztsy70thl9k4l4crk": [],
        #         "tb1qg33wumlly0mn33s776n8njqek3l2rmy8d97r02": [],
        #         "tb1qh48nqrg595g0vfqcw46kt8aun63t9puap5uww5": [],
        #         "tb1qhlgkkqtdglq64xa8wnwz4446mzj97g0jq9ck65": [],
        #         "tb1qjrd86fxf5chezp2pkjzv8v45rzxe8japwxatt2": [],
        #         "tb1qlty269zlah8h9afx0cxvn05qrzyndvnz28ae6d": [],
        #         "tb1qmcnxezw6qh8vys5sxpwfcqgs2t3gyuy7setdky": [],
        #         "tb1qphhms4gdrfpxh5ed2zekhvnwrk84jk5z56qj82": [],
        #         "tb1qq5huaklp886yyqmxju47xm42dgn5demx5yd848": [],
        #         "tb1qqfp0y5qtf0cx7hzat7r3g69gxx96fkg3vnu4cz": [],
        #         "tb1qqu0p6m8ux64kkflx9wa5sc8k39v8crfsn96qky": [],
        #         "tb1qtu73qp3eneht3rfauwz3rttvs7m7l5lahkq66m": [],
        #         "tb1qtvca6a7gp76km7k5kyg7e6lxhvx3qme6x9p53a": [],
        #         "tb1qtxh9dqg07p32e93vpzhfmv7sgp9je5phps5gpy": [],
        #         "tb1qu0wcfv6n96xm6nx5lczcc4gc5h3fjmsfqsmaar": [],
        #         "tb1qx0ld4mml0hkrwuejmetl77rma08ehc4d75qruy": [],
        #         "tb1qxf7jgjd8k0nmm8g8vclx9zg5drr3hfls3p8c6m": [
        #             [
        #                 "23196c7657b60d1fde716c789881ffeee93554b9040c200589c65ceeac051cfb",
        #                 2035683
        #             ],
        #             [
        #                 "c9a54c381293a387a8c1c90e542d11cccf42cd0aeab20be3d1fcd10b465dcfde",
        #                 0
        #             ]
        #         ],
        #         "tb1qy0pklnhg7rks54gzlu003muyllym3xflkd8e7m": [],
        #         "tb1qyt9mkqgf5gfj0lxhmuzv4xamnra8jlxyhpr5x9": [],
        #         "tb1qz9fe873jvrwd0clcknjgj8z6t3uylc50lgvh2q": []
        #     },
        #     "addresses": {
        #         "change": [
        #             "tb1qxf7jgjd8k0nmm8g8vclx9zg5drr3hfls3p8c6m",
        #             "tb1qcx2dultr5x9dc6jz0ye3wq240xfsdxnnl7vpsm",
        #             "tb1q3p86mshyyrmzp2mw50eqeefxtkz8qfl8tsfgrs",
        #             "tb1qe686y2y39ncnafpzxxuyfztsy70thl9k4l4crk",
        #             "tb1qlty269zlah8h9afx0cxvn05qrzyndvnz28ae6d",
        #             "tb1q56f5wgc7wqeqygpfgua64gucafp3wp6293ggcg",
        #             "tb1qqfp0y5qtf0cx7hzat7r3g69gxx96fkg3vnu4cz",
        #             "tb1qhlgkkqtdglq64xa8wnwz4446mzj97g0jq9ck65",
        #             "tb1q4k59fgwhwey8k4rs2l80p73dwtza0z0gvpjjfc",
        #             "tb1qq5huaklp886yyqmxju47xm42dgn5demx5yd848"
        #         ],
        #         "receiving": [
        #             "tb1q46u2rm20rp9rzauxxhdefregggag8cpelh0sqa",
        #             "tb1q8uysvaz8dysekwws868knvxngh5qfjzrlxqf2j",
        #             "tb1qtvca6a7gp76km7k5kyg7e6lxhvx3qme6x9p53a",
        #             "tb1q8ydkp6qhq79ktwpcwjgt9dl7eer32vpw8gvduy",
        #             "tb1qtu73qp3eneht3rfauwz3rttvs7m7l5lahkq66m",
        #             "tb1qu0wcfv6n96xm6nx5lczcc4gc5h3fjmsfqsmaar",
        #             "tb1qx0ld4mml0hkrwuejmetl77rma08ehc4d75qruy",
        #             "tb1qazyz882jeruqlppfralemm3r07pw6cpk426fhw",
        #             "tb1qz9fe873jvrwd0clcknjgj8z6t3uylc50lgvh2q",
        #             "tb1qy0pklnhg7rks54gzlu003muyllym3xflkd8e7m",
        #             "tb1qh48nqrg595g0vfqcw46kt8aun63t9puap5uww5",
        #             "tb1qjrd86fxf5chezp2pkjzv8v45rzxe8japwxatt2",
        #             "tb1qg33wumlly0mn33s776n8njqek3l2rmy8d97r02",
        #             "tb1qmcnxezw6qh8vys5sxpwfcqgs2t3gyuy7setdky",
        #             "tb1q2zqu0rxg70scs9umwwkjxvq7zdcle2ny6pda8v",
        #             "tb1qyt9mkqgf5gfj0lxhmuzv4xamnra8jlxyhpr5x9",
        #             "tb1qphhms4gdrfpxh5ed2zekhvnwrk84jk5z56qj82",
        #             "tb1qqu0p6m8ux64kkflx9wa5sc8k39v8crfsn96qky",
        #             "tb1qch2j8ks2qt6thtf7wdpk6v9gquu8tfca6lsh0n",
        #             "tb1qtxh9dqg07p32e93vpzhfmv7sgp9je5phps5gpy"
        #         ]
        #     },
        #     "fiat_value": {},
        #     "frozen_coins": {},
        #     "imported_channel_backups": {},
        #     "keystore": {
        #         "derivation": "m/0'",
        #         "pw_hash_version": 1,
        #         "root_fingerprint": "8e0c95ad",
        #         "seed": "toss despair unaware moral shield process morning cook enlist park talent snack",
        #         "seed_type": "segwit",
        #         "type": "bip32",
        #         "xprv": "...",
        #         "xpub": "vpub5VGXXixD2pHLFtcKtCF57e8mx2JW6fie8VydXijC8sRKAL4RshgjEmzbmV915NeVB9pd23DVYem6zWM7HXFLNwaffNVHowdD9SJWwESyQhp"
        #     },
        #     "labels": {
        #         "14f6bd3dec07c15773b29b4c7f7d3c913197840dfdf567b6b078e224f021b4ac": "this is a txid label",
        #         "c9a54c381293a387a8c1c90e542d11cccf42cd0aeab20be3d1fcd10b465dcfde": "this is an address label",
        #         "tb1q8ydkp6qhq79ktwpcwjgt9dl7eer32vpw8gvduy": "this is an address label"
        #     },
        #     "lightning_payments": {},
        #     "lightning_preimages": {},
        #     "onchain_channel_backups": {},
        #     "qt-console-history": [],
        #     "seed_type": "segwit",
        #     "seed_version": 41,
        #     "stored_height": 2035683,
        #     "submarine_swaps": {},
        #     "use_encryption": false,
        #     "wallet_type": "standard",
        # }
        elif "keystore" in wallet_data:
            wallet_name = wallet_data["keystore"].get("label", "Electrum Wallet")

            if "xpub" in wallet_data["keystore"]:
                wallet_type = cls.wallet_type_by_slip132_xpub(
                    wallet_data["keystore"]["xpub"], is_multisig=False
                )
            else:
                raise Exception(
                    '"xpub" not found in "keystore" in Electrum backup json'
                )
            recv_descriptor = "{}({})".format(
                wallet_type,
                "[{}]{}/0/*".format(
                    wallet_data["keystore"]["derivation"].replace(
                        "m", wallet_data["keystore"]["root_fingerprint"]
                    ),
                    wallet_data["keystore"]["xpub"],
                ),
            )
            cosigners_types = [
                {
                    "type": wallet_data["keystore"].get("hw_type", "other"),
                    "label": wallet_data["keystore"].get("label", "Electrum Wallet"),
                }
            ]
            if "seed" in wallet_data["keystore"]:
                flash(
                    _(
                        "The Electrum wallet contains a seed. The seed will not be imported."
                    ),
                    "warning",
                )
        # Current Specter backups
        else:
            # Newer exports are able to reinitialize device types but stay backwards
            #   compatible with older backups.
            if "devices" in wallet_data:
                cosigners_types = wallet_data["devices"]

            wallet_name = wallet_data.get("label", "Imported Wallet")
            recv_descriptor = wallet_data.get("descriptor", None)

        return (wallet_name, recv_descriptor, cosigners_types)

    @classmethod
    def wallet_type_by_slip132_xpub(cls, xpub, is_multisig=True):
        """
        see: https://github.com/satoshilabs/slips/blob/master/slip-0132.md
        Electrum backups use SLIP-132 but note that other wallets don't make the same
        guarantee.
        """
        if is_multisig:
            if xpub.startswith("xpub") or xpub.startswith("tpub"):
                return "sh"
            elif xpub.startswith("Ypub") or xpub.startswith("Upub"):
                return "sh-wsh"
            elif xpub.startswith("Zpub") or xpub.startswith("Vpub"):
                return "wsh"
        else:
            if xpub.startswith("xpub") or xpub.startswith("tpub"):
                return "pkh"
            elif xpub.startswith("ypub") or xpub.startswith("upub"):
                return "sh-wpkh"
            elif xpub.startswith("zpub") or xpub.startswith("vpub"):
                return "wpkh"
        raise Exception(f"Unhandled xpub type: {xpub}")



"""
Manages the list of transactions for the wallet
"""
from typing import Union
import os
from .specter_error import SpecterError
from .persistence import delete_file, write_csv, read_csv
from .helpers import get_address_from_dict
from embit.transaction import Transaction
from embit.liquid.networks import get_network
from embit import bip32
import json
import math
import logging
from .util.tx import decoderawtransaction
from .util.psbt import SpecterTx, AbstractTxContext, SpecterPSBT

logger = logging.getLogger(__name__)


def parse_arr(v):
    if not isinstance(v, str):
        return v
    try:
        return json.loads(v.replace("'", '"'))
    except:
        return v


class AbstractTxListContext(AbstractTxContext):
    @property
    def rpc(self):
        if hasattr(self, "parent"):
            return self.parent.rpc
        raise NotImplementedError("Implement this!")

    @property
    def chain(self) -> str:
        if hasattr(self, "parent"):
            return self.parent.chain
        raise NotImplementedError("Implement this!")


class TxItem(dict, AbstractTxListContext):
    TransactionCls = Transaction
    columns = [
        "txid",  # str, txid in hex
        "blockhash",  # str, blockhash, None if not confirmed
        "blockheight",  # int, blockheight, None if not confirmed
        "time",  # int (timestamp in seconds), time received
        "blocktime",  # int (timestamp in seconds), time the block was mined
        "bip125-replaceable",  # str ("yes" / "no"), whatever RBF is enabled for the transaction
        "conflicts",  # rbf conflicts, list of txids
        "vsize",
        "category",
        "address",
        "amount",
        "ismine",
    ]
    type_converter = [
        str,
        str,
        int,
        int,
        int,
        str,
        parse_arr,
        int,
        str,
        parse_arr,
        parse_arr,
        bool,
    ]

    def __init__(self, parent, addresses, rawdir, **kwargs):
        self.parent = parent
        self._addresses = addresses
        self.rawdir = rawdir
        # copy
        kwargs = dict(**kwargs)
        # replace with None or convert
        for i, k in enumerate(self.columns):
            v = kwargs.get(k, "")
            kwargs[k] = None if v in ["", None] else self.type_converter[i](v)

        super().__init__(**kwargs)
        self._tx = None
        # if we have hex data
        if "hex" in kwargs:
            self._tx = self.TransactionCls.from_string(kwargs["hex"])
        # conflicts were renamed to walletconflicts
        if "walletconflicts" in kwargs:
            self["conflicts"] = kwargs["walletconflicts"]

    @property
    def fname(self):
        return os.path.join(self.rawdir, self.txid + ".bin")

    @property
    def tx(self):
        if not self._tx:
            # Get transaction from file if we don't have it cached.
            # We cache transactions to self._tx
            # only when new tx is added before dump() is called
            try:
                if os.path.isfile(self.fname):
                    with open(self.fname, "rb") as f:
                        tx = self.TransactionCls.read_from(f)
                        self._tx = tx
                        return tx
            except Exception as e:
                logger.error(e)
        # if we failed to load tx from file - load it from RPC
        if not self._tx:
            # get transaction from rpc
            try:
                res = self.rpc.gettransaction(self.txid)
                tx = self.TransactionCls.from_string(res["hex"])
                self._tx = tx
                return tx
            except Exception as e:
                logger.error(e)
        return self._tx

    @property
    def vsize(self):
        if self.get("vsize"):
            return self["vsize"]
        tx = self.tx
        txsize = len(tx.serialize())
        if tx.is_segwit:
            # tx size - flag - marker - witness
            non_witness_size = (
                txsize - 2 - sum([len(inp.witness.serialize()) for inp in tx.vin])
            )
            witness_size = txsize - non_witness_size
            weight = non_witness_size * 4 + witness_size
            vsize = math.ceil(weight / 4)
        else:
            vsize = txsize
            weight = txsize * 4
        return vsize

    def dump(self):
        """Dumps transaction in binary to the folder if it's not there"""
        # nothing to do if file exists or we don't have binary tx
        if os.path.isfile(self.fname) or not self._tx:
            return
        # Try to create a directory if it's not there
        # and write raw tx to file
        # Can fail if multiple threads create the same dir
        try:
            # create dir if it doesn't exist
            if not os.path.isdir(self.rawdir):
                os.mkdir(self.rawdir)
        except Exception as e:
            logger.error(e)
        try:
            with open(self.fname, "wb") as f:
                self.tx.write_to(f)
        except Exception as e:
            logger.error(e)
        # clear cached tx as we saved the transaction to file
        self._tx = None

    @property
    def txid(self):
        return self["txid"]

    @property
    def hex(self):
        return str(self.tx)

    def __str__(self):
        return self.txid

    def __repr__(self):
        return f"TxItem({str(self)})"

    def __dict__(self):
        return {
            "txid": self["txid"],
            "blockhash": self["blockhash"],
            "blockheight": self["blockheight"],
            "time": self["time"],
            "blocktime": self["blocktime"],
            "conflicts": self["conflicts"],
            "bip125-replaceable": self["bip125-replaceable"],
            "vsize": self["vsize"],
            "category": self["category"],
            "address": self["address"],
            "amount": self["amount"],
            "ismine": self["ismine"],
        }


class TxList(dict, AbstractTxListContext):
    ItemCls = TxItem  # for inheritance
    PSBTCls = SpecterPSBT

    def __init__(self, path, parent, addresses):
        self.parent = parent
        self.path = path
        # folder to store transactions in binary form
        self.rawdir = path.replace(".csv", "_raw")
        self._addresses = addresses
        txs = []
        file_exists = False
        try:
            if os.path.isfile(self.path):
                txs = read_csv(
                    self.path,
                    self.ItemCls,
                    self,
                    self._addresses,
                    self.rawdir,
                )
                for tx in txs:
                    self[tx.txid] = tx
                file_exists = True
        except Exception as e:
            logger.error(e)
        self._file_exists = file_exists

    def _save(self):
        # check if we have at least one transaction
        if self:
            # Dump all transactions to binary files
            # This happens only if they have not been dumped before
            for tx in self.values():
                tx.dump()
            write_csv(self.path, list(self.values()), self.ItemCls)
            self._file_exists = True
        else:
            delete_file(self.path)
            self._file_exists = False

    def getfetch(self, txid):
        """
        Returns TxItem instance if it is known,
        otherwise tries to get it from rpc, adds to self and returns TxItem
        """
        if txid not in self:
            tx = self.rpc.gettransaction(txid)
            if "time" not in tx:
                tx["time"] = tx["timereceived"]
            self.add({txid: tx})
        return self[txid]

    def gettransaction(self, txid, blockheight=None, decode=False, full=True):
        """
        Will ask Bitcoin Core for a transaction if blockheight is None or txid not known
        Provide blockheight or 0 if you don't care about confirmations number
        to avoid RPC calls.
        full=True will add a "hex" key
        decode=True will decode transaction similar to Core's decoderawtransaction
        """
        # if we don't know blockheigth or transaction
        # we invalidate which results in asking core
        if txid in self and self[txid]["blockheight"] == None:
            self.invalidate(txid)

        if blockheight is None or txid not in self:
            tx = self.rpc.gettransaction(txid)
            if "time" not in tx:
                tx["time"] = tx["timereceived"]
            # save if we don't know about this tx
            if txid not in self:
                self.add({txid: tx})
        # lookup tx in cache, it should be there already
        tx = self[txid]
        res = dict(**tx)
        if full:
            res["hex"] = tx.hex
        if decode:
            res.update(self._decoderawtransaction(tx.hex))
        return res

    def invalidate(self, txid):
        """removes a tx from the list"""
        if txid not in self:
            raise SpecterError(f"TX with txid {txid} does not exit in {self}")
        del self[txid]
        self._save()

    def _decoderawtransaction(self, tx: Union[Transaction, str, bytes]):
        return SpecterTx(self, tx).to_dict()

    def add(self, txs):
        """
        adds transactions to the list without any rpc calls.
        assuming tx is already watched by the wallet.
        Transactions should be a dict with dicts with fields:
        "<txid>": {
            "txid", - hex txid
            "hex",  - hex raw transaction
            "blockheight", - int blockheight if confirmed, None otherwise
            "blockhash", - str blockhash if confirmed, None otherwise
            "time", - int unix timestamp in seconds when tx was received
            "blocktime",  - int unix timestamp in seconds  the block was mined
            "conflicts", - list of txids spending the same inputs (rbf)
            "bip125-replaceable", - str ("yes" or "no") - is rbf enabled for this tx
        }
        """
        # here we store all addresses in transactions
        # to set them used later
        addresses = []
        # first we add all transactions to cache
        for txid in txs:
            tx = txs[txid]
            # find minimal from 3 times:
            maxtime = 10445238000  # TODO: change after 31 dec 2300 lol
            time = min(
                tx.get("blocktime", maxtime),
                tx.get("timereceived", maxtime),
                tx.get("time", maxtime),
            )
            obj = {
                "txid": txid,
                "blockheight": tx.get("blockheight", None),
                "blockhash": tx.get("blockhash", None),
                "time": time,
                "blocktime": tx.get("blocktime", None),
                "conflicts": tx.get("walletconflicts", []),
                "bip125-replaceable": tx.get("bip125-replaceable", "no"),
                "hex": tx.get("hex", None),
            }
            txitem = self.ItemCls(self, self._addresses, self.rawdir, **obj)
            self[txid] = txitem
            if txitem.tx:
                for vout in txitem.tx.vout:
                    try:
                        addr = vout.script_pubkey.address(get_network(self.chain))
                        if addr not in addresses:
                            addresses.append(addr)
                    except:
                        pass  # maybe not an address, but a raw script?
        self._addresses.set_used(addresses)
        # detect category, amounts and addresses
        for tx in [self[txid] for txid in self if txid in txs]:
            self._fill_missing(tx)
        self._save()

    def _update_destinations(self, tx, outs):
        addresses = [out.get("address", "Unknown") for out in outs]
        amounts = [out["float_amount"] for out in outs]
        if len(addresses) == 1:
            addresses = addresses[0]
            amounts = amounts[0]
        tx["address"] = addresses
        tx["amount"] = amounts

    def _get_psbt(self, raw_tx):
        psbt = self.PSBTCls.from_transaction(raw_tx, self.descriptor, self.network)
        # fill derivation paths etc
        updated = self.rpc.walletprocesspsbt(str(psbt), False).get("psbt", None)
        if updated:
            psbt.update(updated)
        return psbt

    def _fill_missing(self, tx):
        """This seem to calculate the category of the tx which is one of:
        mixed (default), generate, selftransfer, receive or send

        Also the tx gets a key with a boolean to figure out whether its "mine"
        """
        raw_tx = tx.tx
        psbt = self._get_psbt(raw_tx)
        # detect category
        category = "mixed"

        # calculate everything once
        inputs = [inp.to_dict() for inp in psbt.inputs]
        outputs = [out.to_dict() for out in psbt.outputs]
        all_inputs_mine = all([inp["is_mine"] for inp in inputs])
        all_outputs_mine = all([out["is_mine"] for out in outputs])
        all_inputs_external = not any([inp["is_mine"] for inp in inputs])

        if b"\x00" * 32 in [vin.txid for vin in raw_tx.vin]:
            category = "generate"
        elif all_inputs_mine and all_outputs_mine:
            category = "selftransfer"
        elif all_inputs_external:
            category = "receive"
        elif all_inputs_mine:
            category = "send"

        all_outs = [out for out in outputs]
        my_outs = [out for out in all_outs if out["is_mine"]]
        my_receiving = [out for out in my_outs if not out["change"]]
        external = [out for out in all_outs if out not in my_outs]
        # decide what addresses to show
        if category in ["generate", "receive", "selftransfer"]:
            # either receiving only (if not empty), or only mine (if not empty), or all
            outs = my_receiving or my_outs or all_outs
        elif category in ["send"]:
            # keep only external addresses if they are present
            outs = external or all_outs
        else:
            # not sure what's the best here
            outs = my_receiving or my_outs or external or all_outs

        self._update_destinations(tx, outs)
        tx["category"] = category
        # at least one input or output is ours - tx is ours
        tx["ismine"] = any(scope["is_mine"] for scope in (inputs + outputs))

    def load(self, arr):
        """
        TODO: load transactions from backup
        Load transactions to Core with merkle proofs to avoid rescan
        arr should be a dict with dicts:
        "<txid>": {
            "txid",         - hex txid
            "hex",          - optional, raw hex transactions
            "merkle_proof", - optional, hex merkle proof
            "blockheight",  - optional, block height
            "blockhash",    - optional, block hash
            "time",         - optional, time received
        }
        If optional fields are not found we will try to get them from rpc calls.
        If we fail - transaction will not be added.
        Retuns an dict of dicts:
        {
            <txid>: {"success": True}, on success,
            <txid>: {"success": False, "error": "reason"}, if failed to import transaction
        }
        """
        # TODO: how to handle unconfirmed?
        # try to get raw transactions and tx details if not present
        # self.save()
        return [{"success": False, "error": "not implemented"} for a in arr]

    @property
    def file_exists(self):
        return self._file_exists and os.path.isfile(self.path)



from flask_babel import Babel, get_translations
from jinja2.utils import markupsafe
import html


class HTMLSafeBabel(Babel):
    """
    A subclass of Flash-Babel which automatically HTML escapes gettext strings

    Jinja2 treats gettext() values as HTML safe by default. This function wraps
    the gettext functions to explictly HTML escape the translated strings before
    they are used in a template.
    """

    def init_app(self, app, *args, **kwargs):
        super(HTMLSafeBabel, self).init_app(app, *args, **kwargs)
        if self._configure_jinja:
            app.jinja_env.install_gettext_callables(
                lambda x: self.reescape_string(get_translations().ugettext(x)),
                lambda s, p, n: self.reescape_string(
                    get_translations().ungettext(s, p, n)
                ),
                newstyle=True,
            )

    @staticmethod
    def reescape_string(escaped_string):
        """
        Decode HTML entities and re-encoded safely using the Jinja HTML escaper.

        Existing entities need to be decoded first to avoid double-encoding.
        """
        return markupsafe.Markup.escape(html.unescape(escaped_string))



import logging
from ..specter_migrator import SpecterMigration

logger = logging.getLogger(__name__)


class SpecterMigration_0000(SpecterMigration):
    # the version that the stuff which is suppose to be migrated has been introduced
    # If this instance has been started AFTER this version, then this migration will never
    # get executed ( see the default-implementation of SpecterMigration.should_execute())
    version = "v1.5.0"  # Faking it here for the unit-tests

    def __init__(self, specter_migrator):
        self.specter_migrator = specter_migrator

    def execute(self):
        logger.debug(
            f"migration_0000 is an example method showing how to implement a migration"
        )

    @property
    def description(self) -> str:
        """Should return a (multiline) description of the migration which will get log.info() at execution-time"""
        return """A dummy migration:
            * It will do nothing
            * It's just here to explain how SpecterMigration works
            * It will be shown in the logs when it's executed (just like the other real migrations but doing nothing)
        """



"""
Blockstream Jade Devices
************************
"""

from .jadepy import jade
from .jadepy.jade import JadeAPI, JadeError
from serial.tools import list_ports

from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union
from hwilib.descriptor import PubkeyProvider, MultisigDescriptor
from hwilib.hwwclient import HardwareWalletClient
from hwilib.errors import (
    ActionCanceledError,
    BadArgumentError,
    DeviceConnectionError,
    DeviceFailureError,
    UnavailableActionError,
    common_err_msgs,
    handle_errors,
)
from hwilib.common import (
    AddressType,
    Chain,
)
from hwilib.key import ExtendedKey, parse_path, KeyOriginInfo, is_hardened
from hwilib.psbt import PSBT
from hwilib.tx import CTransaction
from hwilib._script import (
    is_p2sh,
    is_p2wpkh,
    is_p2wsh,
    is_witness,
    parse_multisig,
)

import logging
import os
import hashlib

# embit-related things
from embit import ec, bip32
from embit.psbt import PSBT
from embit import script
from embit.liquid.pset import PSET
from embit import hashes
from embit.util import secp256k1
from embit.liquid.finalizer import finalize_psbt
from embit.liquid.transaction import write_commitment

# The test emulator port
SIMULATOR_PATH = "tcp:127.0.0.1:2222"

JADE_DEVICE_IDS = [(0x10C4, 0xEA60), (0x1A86, 0x55D4)]
HAS_NETWORKING = hasattr(jade, "_http_request")

py_enumerate = (
    enumerate  # To use the enumerate built-in, since the name is overridden below
)

logger = logging.getLogger(__name__)


def jade_exception(f: Callable[..., Any]) -> Any:
    @wraps(f)
    def func(*args: Any, **kwargs: Any) -> Any:
        try:
            return f(*args, **kwargs)
        except ValueError as e:
            raise BadArgumentError(str(e))
        except JadeError as e:
            if e.code == JadeError.USER_CANCELLED:
                raise ActionCanceledError(f"{f.__name__} canceled by user")
            elif e.code == JadeError.BAD_PARAMETERS:
                raise BadArgumentError(e.message)
            elif e.code == JadeError.INTERNAL_ERROR:
                raise DeviceFailureError(e.message)
            elif e.code == JadeError.HW_LOCKED:
                raise DeviceConnectionError("Device is locked")
            elif e.code == JadeError.NETWORK_MISMATCH:
                raise DeviceConnectionError("Network/chain selection error")
            elif e.code in [
                JadeError.INVALID_REQUEST,
                JadeError.UNKNOWN_METHOD,
                JadeError.PROTOCOL_ERROR,
            ]:
                raise DeviceConnectionError("Messaging/communiciation error")
            else:
                raise e

    return func


# This class extends the HardwareWalletClient for Blockstream Jade specific things
class JadeClient(HardwareWalletClient):

    NETWORKS = {
        Chain.MAIN: "mainnet",
        Chain.TEST: "testnet",
        Chain.REGTEST: "localtest",
    }
    liquid_network = None

    def set_liquid_network(self, chain):
        if chain == "liquidv1":
            self.liquid_network = "liquid"
        elif chain == "liquidtestnet":
            self.liquid_network = "testnet-liquid"
        else:
            self.liquid_network = "localtest-liquid"

    def _network(self):
        if self.liquid_network:
            return self.liquid_network
        return JadeClient.NETWORKS.get(self.chain, "mainnet")

    ADDRTYPES = {
        AddressType.LEGACY: "pkh(k)",
        AddressType.WIT: "wpkh(k)",
        AddressType.SH_WIT: "sh(wpkh(k))",
    }
    MULTI_ADDRTYPES = {
        AddressType.LEGACY: "sh(multi(k))",
        AddressType.WIT: "wsh(multi(k))",
        AddressType.SH_WIT: "sh(wsh(multi(k)))",
    }

    @staticmethod
    def _convertAddrType(addrType, multisig=False):
        if multisig:
            return JadeClient.MULTI_ADDRTYPES[addrType]
        return JadeClient.ADDRTYPES[addrType]

    @staticmethod
    def _get_multisig_name(
        type: str, threshold: int, signers: List[Tuple[bytes, Sequence[int]]]
    ) -> str:
        # Concatenate script-type, threshold, and all signers fingerprints and derivation paths (sorted)
        summary = type + "|" + str(threshold) + "|"
        for fingerprint, path in sorted(signers):
            summary += fingerprint.hex() + "|" + str(path) + "|"

        # Hash it, get the first 6-bytes as hex, prepend with 'hwi'
        hash_summary = hashlib.sha256(summary.encode()).digest().hex()
        return "hwi" + hash_summary[:12]

    def __init__(self, path: str, password: str = "", expert: bool = False) -> None:
        super(JadeClient, self).__init__(path, password, expert)
        self.jade = JadeAPI.create_serial(path)
        self.jade.connect()

        # Push some host entropy into jade
        self.jade.add_entropy(os.urandom(32))

        # Do the PIN thing if required
        # NOTE: uses standard 'requests' networking to connect to blind pinserver
        try:
            while not self.jade.auth_user(self._network()):
                logging.debug("Incorrect PIN provided")
        except:
            try:
                self.chain = Chain.TEST
                while not self.jade.auth_user(self._network()):
                    logging.debug("Incorrect PIN provided")
            except:
                self.chain = Chain.REGTEST
                while not self.jade.auth_user(self._network()):
                    logging.debug("Incorrect PIN provided")

    # Retrieves the public key at the specified BIP 32 derivation path
    @jade_exception
    def get_pubkey_at_path(self, bip32_path: str) -> ExtendedKey:
        path = parse_path(bip32_path)
        xpub = self.jade.get_xpub(self._network(), path)
        ext_key = ExtendedKey.deserialize(xpub)
        return ext_key

    @jade_exception
    def get_master_blinding_key(self) -> str:
        mbk = self.jade.get_master_blinding_key()
        assert len(mbk) == 32
        bkey = ec.PrivateKey(mbk)
        return bkey.wif()

    # Walk the PSBT looking for inputs we can sign.  Push any signatures into the
    # 'partial_sigs' map in the input, and return the updated PSBT.
    @jade_exception
    def sign_tx(self, tx: PSBT) -> PSBT:
        """
        Sign a transaction with the Blockstream Jade.
        """
        # Helper to get multisig record for change output
        def _parse_signers(
            hd_keypath_origins: List[KeyOriginInfo],
        ) -> Tuple[List[Tuple[bytes, Sequence[int]]], List[Sequence[int]]]:
            # Split the path at the last hardened path element
            def _split_at_last_hardened_element(
                path: Sequence[int],
            ) -> Tuple[Sequence[int], Sequence[int]]:
                for i in range(len(path), 0, -1):
                    if is_hardened(path[i - 1]):
                        return (path[:i], path[i:])
                return ([], path)

            signers = []
            paths = []
            for origin in hd_keypath_origins:
                prefix, suffix = _split_at_last_hardened_element(origin.path)
                signers.append((origin.fingerprint, prefix))
                paths.append(suffix)
            # sort signers and paths like in multisig registration
            signers, paths = [list(a) for a in zip(*sorted(zip(signers, paths)))]

            return signers, paths

        c_txn = CTransaction(tx.tx)
        master_fp = self.get_master_fingerprint()
        signing_singlesigs = False
        signing_multisigs = {}
        need_to_sign = True

        while need_to_sign:
            signing_pubkeys: List[Optional[bytes]] = [None] * len(tx.inputs)
            need_to_sign = False

            # Signing input details
            jade_inputs = []
            for n_vin, (txin, psbtin) in py_enumerate(zip(c_txn.vin, tx.inputs)):
                # Get bip32 path to use to sign, if required for this input
                path = None
                multisig_input = len(psbtin.hd_keypaths) > 1
                for pubkey, origin in psbtin.hd_keypaths.items():
                    if origin.fingerprint == master_fp and len(origin.path) > 0:
                        if not multisig_input:
                            signing_singlesigs = True

                        if psbtin.partial_sigs.get(pubkey, None) is None:
                            # hw to sign this input - it is not already signed
                            if signing_pubkeys[n_vin] is None:
                                signing_pubkeys[n_vin] = pubkey
                                path = origin.path
                            else:
                                # Additional signature needed for this input - ie. a multisig where this wallet is
                                # multiple signers?  Clumsy, but just loop and go through the signing procedure again.
                                need_to_sign = True

                # Get the tx and prevout/scriptcode
                utxo = None
                p2sh = False
                input_txn_bytes = None
                if psbtin.witness_utxo:
                    utxo = psbtin.witness_utxo
                if psbtin.non_witness_utxo:
                    utxo = psbtin.non_witness_utxo.vout[txin.prevout.n]
                    input_txn_bytes = (
                        psbtin.non_witness_utxo.serialize_without_witness()
                    )
                if utxo is None:
                    raise Exception(
                        "PSBT is missing input utxo information, cannot sign"
                    )
                scriptcode = utxo.scriptPubKey

                if is_p2sh(scriptcode):
                    scriptcode = psbtin.redeem_script
                    p2sh = True

                witness_input, witness_version, witness_program = is_witness(scriptcode)

                if witness_input:
                    if is_p2wsh(scriptcode):
                        scriptcode = psbtin.witness_script
                    elif is_p2wpkh(scriptcode):
                        scriptcode = b"\x76\xa9\x14" + witness_program + b"\x88\xac"
                    else:
                        continue

                # If we are signing a multisig input, deduce the potential
                # registration details and cache as a potential change wallet
                if multisig_input and path and scriptcode and (p2sh or witness_input):
                    parsed = parse_multisig(scriptcode)
                    if parsed:
                        addr_type = (
                            AddressType.LEGACY
                            if not witness_input
                            else AddressType.WIT
                            if not p2sh
                            else AddressType.SH_WIT
                        )
                        script_variant = self._convertAddrType(addr_type, multisig=True)
                        threshold = parsed[0]

                        pubkeys = parsed[1]
                        hd_keypath_origins = [
                            psbtin.hd_keypaths[pubkey] for pubkey in pubkeys
                        ]

                        signers, paths = _parse_signers(hd_keypath_origins)

                        multisig_name = self._get_multisig_name(
                            script_variant, threshold, signers
                        )
                        signing_multisigs[multisig_name] = (
                            script_variant,
                            threshold,
                            signers,
                        )

                # Build the input and add to the list - include some host entropy for AE sigs (although we won't verify)
                jade_inputs.append(
                    {
                        "is_witness": witness_input,
                        "input_tx": input_txn_bytes,
                        "script": scriptcode,
                        "path": path,
                        "ae_host_entropy": os.urandom(32),
                        "ae_host_commitment": os.urandom(32),
                    }
                )

            # Change output details
            # This is optional, in that if we send it Jade validates the change output script
            # and the user need not confirm that ouptut.  If not passed the change output must
            # be confirmed by the user on the hwwallet screen, like any other spend output.
            change: List[Optional[Dict[str, Any]]] = [None] * len(tx.outputs)

            # If signing multisig inputs, get registered multisigs details in case we
            # see any multisig outputs which may be change which we can auto-validate.
            # ie. filter speculative 'signing multisigs' to ones actually registered on the hw
            candidate_multisigs = {}

            if signing_multisigs:
                # register multisig if xpubs are known
                if tx.xpub and len(signing_multisigs) == 1:
                    msigname = list(signing_multisigs.keys())[0]
                    signers = []
                    origins = []
                    for xpub in tx.xpub:
                        hd = bip32.HDKey.parse(xpub)
                        origin = tx.xpub[xpub]
                        origins.append((origin.fingerprint, origin.path))

                        signers.append(
                            {
                                "fingerprint": origin.fingerprint,
                                "derivation": origin.path,
                                "xpub": str(hd),
                                "path": [],
                            }
                        )

                    # sort origins and signers together
                    origins, signers = [
                        list(a) for a in zip(*sorted(zip(origins, signers)))
                    ]

                    # Get a deterministic name for this multisig wallet
                    script_variant = signing_multisigs[msigname][0]
                    thresh = signing_multisigs[msigname][1]
                    num_signers = signing_multisigs[msigname][2]
                    multisig_name = self._get_multisig_name(
                        script_variant, thresh, origins
                    )
                    # stupid sanity check of the fingerprints and origins
                    if multisig_name == msigname:
                        # Need to ensure this multisig wallet is registered first
                        # (Note: 're-registering' is a no-op)
                        self.jade.register_multisig(
                            self._network(),
                            multisig_name,
                            script_variant,
                            True,  # always use sorted
                            thresh,
                            signers,
                        )
                #
                registered_multisigs = self.jade.get_registered_multisigs()
                signing_multisigs = {
                    k: v
                    for k, v in signing_multisigs.items()
                    if k in registered_multisigs
                    and registered_multisigs[k]["variant"] == v[0]
                    and registered_multisigs[k]["threshold"] == v[1]
                    and registered_multisigs[k]["num_signers"] == len(v[2])
                }

            # Look at every output...
            for n_vout, (txout, psbtout) in py_enumerate(zip(c_txn.vout, tx.outputs)):
                num_signers = len(psbtout.hd_keypaths)

                if num_signers == 1 and signing_singlesigs:
                    # Single-sig output - since we signed singlesig inputs this could be our change
                    for pubkey, origin in psbtout.hd_keypaths.items():
                        # Considers 'our' outputs as potential change as far as Jade is concerned
                        # ie. can be verified and auto-confirmed.
                        # Is this ok, or should check path also, assuming bip44-like ?
                        if origin.fingerprint == master_fp and len(origin.path) > 0:
                            change_addr_type = None
                            if txout.is_p2pkh():
                                change_addr_type = AddressType.LEGACY
                            elif txout.is_witness()[0] and not txout.is_p2wsh():
                                change_addr_type = AddressType.WIT  # ie. p2wpkh
                            elif (
                                txout.is_p2sh() and is_witness(psbtout.redeem_script)[0]
                            ):
                                change_addr_type = AddressType.SH_WIT
                            else:
                                continue

                            script_variant = self._convertAddrType(
                                change_addr_type, multisig=False
                            )
                            change[n_vout] = {
                                "path": origin.path,
                                "variant": script_variant,
                            }

                elif num_signers > 1 and signing_multisigs:
                    # Multisig output - since we signed multisig inputs this could be our change
                    candidate_multisigs = {
                        k: v
                        for k, v in signing_multisigs.items()
                        if len(v[2]) == num_signers
                    }
                    if not candidate_multisigs:
                        continue

                    for pubkey, origin in psbtout.hd_keypaths.items():
                        if origin.fingerprint == master_fp and len(origin.path) > 0:
                            change_addr_type = None
                            if (
                                txout.is_p2sh()
                                and not is_witness(psbtout.redeem_script)[0]
                            ):
                                change_addr_type = AddressType.LEGACY
                                scriptcode = psbtout.redeem_script
                            elif txout.is_p2wsh() and not txout.is_p2sh():
                                change_addr_type = AddressType.WIT
                                scriptcode = psbtout.witness_script
                            elif (
                                txout.is_p2sh() and is_witness(psbtout.redeem_script)[0]
                            ):
                                change_addr_type = AddressType.SH_WIT
                                scriptcode = psbtout.witness_script
                            else:
                                continue

                            parsed = parse_multisig(scriptcode)
                            if parsed:
                                script_variant = self._convertAddrType(
                                    change_addr_type, multisig=True
                                )
                                threshold = parsed[0]

                                pubkeys = parsed[1]
                                hd_keypath_origins = [
                                    psbtout.hd_keypaths[pubkey] for pubkey in pubkeys
                                ]

                                signers, paths = _parse_signers(hd_keypath_origins)

                                multisig_name = self._get_multisig_name(
                                    script_variant, threshold, signers
                                )

                                matched_multisig = candidate_multisigs.get(
                                    multisig_name
                                ) == (script_variant, threshold, signers)
                                if matched_multisig:
                                    change[n_vout] = {
                                        "paths": paths,
                                        "multisig_name": multisig_name,
                                    }

            # The txn itself
            txn_bytes = c_txn.serialize_without_witness()

            # Request Jade generate the signatures for our inputs.
            # Change details are passed to be validated on the hw (user does not confirm)
            signatures = self.jade.sign_tx(
                self._network(), txn_bytes, jade_inputs, change, True
            )

            # Push sigs into PSBT structure as appropriate
            for psbtin, signer_pubkey, sigdata in zip(
                tx.inputs, signing_pubkeys, signatures
            ):
                signer_commitment, sig = sigdata
                if signer_pubkey and sig:
                    psbtin.partial_sigs[signer_pubkey] = sig

        # Return the updated psbt
        return tx

    # Sign message, confirmed on device
    @jade_exception
    def sign_message(self, message: Union[str, bytes], bip32_path: str) -> str:
        path = parse_path(bip32_path)
        if isinstance(message, bytes) or isinstance(message, bytearray):
            message = message.decode("utf-8")
        signature = self.jade.sign_message(path, message)
        return signature

    # Display address of specified type on the device. Only supports single-key based addresses atm.
    @jade_exception
    def display_singlesig_address(self, bip32_path: str, addr_type: AddressType) -> str:
        path = parse_path(bip32_path)
        addr_type = self._convertAddrType(addr_type)
        address = self.jade.get_receive_address(
            self._network(), path, variant=addr_type
        )
        return address

    def display_multisig_address(
        self,
        addr_type: AddressType,
        multisig: MultisigDescriptor,
    ) -> str:
        signer_origins = []
        signers = []
        paths = []
        for pubkey in multisig.pubkeys:
            if pubkey.extkey is None:
                raise BadArgumentError(
                    "Blockstream Jade can only generate addresses for multisigs with full extended keys"
                )
            if pubkey.origin is None:
                raise BadArgumentError(
                    "Blockstream Jade can only generate addresses for multisigs with key origin information"
                )
            if pubkey.deriv_path is None:
                raise BadArgumentError(
                    "Blockstream Jade can only generate addresses for multisigs with key origin derivation path information"
                )

            # Tuple to derive deterministic name for the registrtion
            signer_origins.append((pubkey.origin.fingerprint, pubkey.origin.path))

            #  We won't include the additional path in the multisig registration
            signers.append(
                {
                    "fingerprint": pubkey.origin.fingerprint,
                    "derivation": pubkey.origin.path,
                    "xpub": pubkey.pubkey,
                    "path": [],
                }
            )

            # Instead hold it as the address path
            path = (
                pubkey.deriv_path[1:]
                if pubkey.deriv_path[0] == "/"
                else pubkey.deriv_path
            )
            paths.append(parse_path(path))

        # sort origins, signers and paths according to origins (like in _get_multisig_name)
        signer_origins, signers, paths = [
            list(a) for a in zip(*sorted(zip(signer_origins, signers, paths)))
        ]

        # Get a deterministic name for this multisig wallet
        script_variant = self._convertAddrType(addr_type, multisig=True)
        multisig_name = self._get_multisig_name(
            script_variant, multisig.thresh, signer_origins
        )

        # Need to ensure this multisig wallet is registered first
        # (Note: 're-registering' is a no-op)
        self.jade.register_multisig(
            self._network(),
            multisig_name,
            script_variant,
            True,  # always use sorted
            multisig.thresh,
            signers,
        )
        address = self.jade.get_receive_address(
            self._network(), paths, multisig_name=multisig_name
        )

        return str(address)

    # Setup a new device
    def setup_device(self, label="", passphrase=""):
        """
        The Blockstream Jade does not support setup via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError(
            "The Blockstream Jade does not support software setup"
        )

    # Wipe this device
    def wipe_device(self):
        """
        The Blockstream Jade does not support wiping via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError(
            "The Blockstream Jade does not support wiping via software"
        )

    # Restore device from mnemonic or xprv
    def restore_device(self, label="", word_count=24):
        """
        The Blockstream Jade does not support restoring via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError(
            "The Blockstream Jade does not support restoring via software"
        )

    # Begin backup process
    def backup_device(self, label="", passphrase=""):
        """
        The Blockstream Jade does not support backing up via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError(
            "The Blockstream Jade does not support creating a backup via software"
        )

    # Close the device
    def close(self):
        self.jade.disconnect()

    # Prompt pin
    def prompt_pin(self):
        """
        The Blockstream Jade does not need a PIN sent from the host.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError(
            "The Blockstream Jade does not need a PIN sent from the host"
        )

    # Send pin
    def send_pin(self, pin):
        """
        The Blockstream Jade does not need a PIN sent from the host.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError(
            "The Blockstream Jade does not need a PIN sent from the host"
        )

    # Toggle passphrase
    def toggle_passphrase(self):
        """
        The Blockstream Jade does not support toggling passphrase from the host.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError(
            "The Blockstream Jade does not support toggling passphrase from the host"
        )

    def _blind(self, pset, seed: bytes = None):
        if seed is None:
            seed = pset.unknown.get(b"\xfc\x07specter\x00", os.urandom(32))
        txseed = pset.txseed(seed)
        # assign blinding factors to all outputs
        blinding_outs = []
        commitments = []
        # because we do sha once (cause taproot), and they want sha twice
        hash_prevouts = hashes.sha256(pset.blinded_tx.hash_prevouts())
        last_i = 0
        last_commitment = {}
        for i, out in py_enumerate(pset.outputs):
            # skip ones where we don't need blinding
            if out.blinding_pubkey is None:
                commitments.append(None)
                continue
            commitment = self.jade.get_commitments(
                bytes(reversed(out.asset)), out.value, hash_prevouts, i, vbf=None
            )
            commitment["blinding_key"] = out.blinding_pubkey
            commitments.append(commitment)
            last_i = i
            last_commitment = commitments[-1]
            out.asset_blinding_factor = commitment["abf"]
            out.value_blinding_factor = commitment["vbf"]
            blinding_outs.append(out)
        if len(blinding_outs) == 0:
            raise Exception("Nothing to blind")
        # calculate last vbf
        vals = []
        abfs = []
        vbfs = []
        for sc in pset.inputs + blinding_outs:
            value = sc.value if sc.value is not None else sc.utxo.value
            asset = sc.asset or sc.utxo.asset
            if not (isinstance(value, int) and len(asset) == 32):
                continue
            vals.append(value)
            abfs.append(sc.asset_blinding_factor or b"\x00" * 32)
            vbfs.append(sc.value_blinding_factor or b"\x00" * 32)
        last_vbf = secp256k1.pedersen_blind_generator_blind_sum(
            vals, abfs, vbfs, len(vals) - len(blinding_outs)
        )
        last_out = blinding_outs[-1]
        new_last_commitment = self.jade.get_commitments(
            bytes(reversed(last_out.asset)),
            last_out.value,
            hash_prevouts,
            last_i,
            vbf=last_vbf,
        )
        # check abf didn't change
        assert new_last_commitment["abf"] == last_out.asset_blinding_factor
        # set new values in the last commitment
        last_commitment.update(new_last_commitment)
        blinding_outs[-1].value_blinding_factor = last_vbf

        # calculate commitments (surj proof etc)
        in_tags = []
        in_gens = []
        for inp in pset.inputs:
            if inp.asset:
                in_tags.append(inp.asset)
                in_gens.append(secp256k1.generator_parse(inp.utxo.asset))
            # if we have unconfidential input
            elif len(inp.utxo.asset) == 32:
                in_tags.append(inp.utxo.asset)
                in_gens.append(secp256k1.generator_generate(inp.utxo.asset))

        for i, out in py_enumerate(pset.outputs):
            if None in [out.blinding_pubkey, out.value, out.asset_blinding_factor]:
                continue
            gen = secp256k1.generator_generate_blinded(
                out.asset, out.asset_blinding_factor
            )
            out.asset_commitment = secp256k1.generator_serialize(gen)
            value_commitment = secp256k1.pedersen_commit(
                out.value_blinding_factor, out.value, gen
            )
            out.value_commitment = secp256k1.pedersen_commitment_serialize(
                value_commitment
            )

            proof_seed = hashes.tagged_hash(
                "liquid/surjection_proof", txseed + i.to_bytes(4, "little")
            )
            proof, in_idx = secp256k1.surjectionproof_initialize(
                in_tags, out.asset, proof_seed
            )
            secp256k1.surjectionproof_generate(
                proof, in_idx, in_gens, gen, abfs[in_idx], out.asset_blinding_factor
            )
            out.surjection_proof = secp256k1.surjectionproof_serialize(proof)
            del proof

            # generate range proof
            rangeproof_nonce = hashes.tagged_hash(
                "liquid/range_proof", txseed + i.to_bytes(4, "little")
            )
            # reblind with extra message for unblinding of change outs
            extra_message = (
                out.unknown.get(b"\xfc\x07specter\x01", b"")
                if out.bip32_derivations
                else b""
            )
            out.reblind(
                rangeproof_nonce,
                extra_message=extra_message,
            )

        return commitments

    def sign_pset(self, b64pset: str) -> str:
        """Signs specter-desktop specific Liquid PSET transaction"""
        mfp = self.get_master_fingerprint()
        pset = PSET.from_string(b64pset)
        commitments = self._blind(pset)
        ins = [
            {
                "is_witness": True,
                # "input_tx": inp.non_witness_utxo.serialize(),
                "script": inp.witness_script.data
                if inp.witness_script
                else script.p2pkh_from_p2wpkh(inp.script_pubkey).data,
                "value_commitment": write_commitment(inp.utxo.value),
                "path": [
                    der
                    for der in inp.bip32_derivations.values()
                    if der.fingerprint == mfp
                ][0].derivation,
            }
            for inp in pset.inputs
        ]
        change = [
            {
                "path": [
                    der
                    for pub, der in out.bip32_derivations.items()
                    if der.fingerprint == mfp
                ][0].derivation,
                "variant": self._get_script_type(out),
            }
            if out.bip32_derivations and self._get_script_type(out) is not None
            else None
            for out in pset.outputs
        ]
        rawtx = pset.blinded_tx.serialize()

        signatures = self.jade.sign_liquid_tx(
            self._network(), rawtx, ins, commitments, change
        )
        for i, inp in py_enumerate(pset.inputs):
            inp.partial_sigs[
                [
                    pub
                    for pub, der in inp.bip32_derivations.items()
                    if der.fingerprint == mfp
                ][0]
            ] = signatures[i]
        # we must finalize here because it has different commitments and only supports singlesig
        return str(finalize_psbt(pset))

    def _get_script_type(self, out):
        if out.script_pubkey.script_type() == "p2pkh":
            return "pkh(k)"
        elif out.script_pubkey.script_type() == "p2wpkh":
            return "wpkh(k)"
        elif out.script_pubkey.script_type() == "p2sh":
            if out.redeem_script.script_type() == "p2wpkh":
                return "sh(wpkh(k))"
        # otherwise None
        return None


def enumerate(password: str = "") -> List[Dict[str, Any]]:
    results = []

    def _get_device_entry(device_model: str, device_path: str) -> Dict[str, Any]:
        d_data: Dict[str, Any] = {}
        d_data["type"] = "jade"
        d_data["model"] = device_model
        d_data["path"] = device_path
        d_data["needs_pin_sent"] = False
        d_data["needs_passphrase_sent"] = False

        client = None
        with handle_errors(common_err_msgs["enumerate"], d_data):
            client = JadeClient(device_path, password, timeout=1)
            d_data["fingerprint"] = client.get_master_fingerprint().hex()

        if client:
            client.close()

        return d_data

    # Jade is not really an HID device, it shows as a serial/com port device.
    # Scan com ports looking for the relevant vid and pid, and use 'path' to
    # hold the path to the serial port device, eg. /dev/ttyUSB0
    for devinfo in list_ports.comports():
        if (devinfo.vid, devinfo.pid) in JADE_DEVICE_IDS:
            results.append(_get_device_entry("jade", devinfo.device))

    # If we can connect to the simulator, add it too
    try:
        with JadeAPI.create_serial(SIMULATOR_PATH, timeout=1) as jade:
            verinfo = jade.get_version_info()

        if verinfo is not None:
            results.append(_get_device_entry("jade_simulator", SIMULATOR_PATH))

    except Exception as e:
        # If we get any sort of error do not add the simulator
        logger.debug(f"Failed to connect to Jade simulator at {SIMULATOR_PATH}")
        logger.debug(e)

    return results



""" API Backend - Base Resource Models """

from functools import wraps
import logging
import re
from flask_restful import Resource, abort

from cryptoadvance.specter.api import api_rest
from cryptoadvance.specter.api.security import require_admin
from cryptoadvance.specter.api import auth
from cryptoadvance.specter.specter_error import SpecterError

logger = logging.getLogger(__name__)


def error_handling(func):
    """User needs Admin-rights method decorator"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        """wrapping the whole error-handling around methods"""
        try:
            return func(*args, **kwargs)
        except SpecterError as se:
            logger.error(se)
            # Not that elegant as this function is accumulating all the error-handling for different
            # endpoints. On the other hand, this is probabyl more tied to different SpecterErrors
            # rather than the implementation of a rest-endpoint.
            match = re.match("Wallet (.+) does not exist!", str(se))
            if match:
                return abort(403, message=f"Wallet {match.group(1)} does not exist")
            if str(se).endswith(
                "does not have sufficient funds to make the transaction."
            ):
                return abort(
                    412,
                    message=f"Wallet does not have sufficient funds to make the transaction.",
                )
            return abort(500)
        except Exception as e:
            logger.error("Unexpected Exception in Rest-Request:")
            logger.exception(e)
            return abort(
                500,
                message="Can't tell you the reason of the issue. Please check the logs",
            )

    return wrapper


class BaseResource(Resource):
    """A baseClass for rsources which returns Method not allowed by default"""

    method_decorators = [error_handling]

    def get(self, *args, **kwargs):
        abort(405)

    def post(self, *args, **kwargs):
        abort(405)

    def put(self, *args, **kwargs):
        abort(405)

    def patch(self, *args, **kwargs):
        abort(405)

    def delete(self, *args, **kwargs):
        abort(405)


class SecureResource(BaseResource):
    """A REST-resource which makes sure that the user is Authenticated"""

    method_decorators = [error_handling, auth.login_required]


class AdminResource(BaseResource):
    """A REST-resource which makes sure that the user is an admin"""

    method_decorators = [error_handling, require_admin, auth.login_required]


def rest_resource(resource_cls):
    """Decorator for adding resources to Api App"""
    api_rest.add_resource(resource_cls, *resource_cls.endpoints)



"""
Hardware Wallet Client Interface
********************************

The :class:`SpecterClient` is the class to communicate with Specter-DIY hardware wallet.
"""

import serial
import serial.tools.list_ports
import socket, time
from hwilib.hwwclient import *
from hwilib.errors import (
    ActionCanceledError,
    BadArgumentError,
    DeviceBusyError,
    DeviceFailureError,
    UnavailableActionError,
)
import logging
import hashlib
from binascii import a2b_base64, b2a_base64

py_enumerate = enumerate
logger = logging.getLogger(__name__)


class SpecterClient(HardwareWalletClient):
    """Create a client for a HID device that has already been opened.

    This abstract class defines the methods
    that hardware wallet subclasses should implement.
    """

    # timeout large enough to handle xpub derivations
    TIMEOUT = 3

    def __init__(self, path: str, password: str = "", expert: bool = False) -> None:
        """
        :param path: Path to the device as returned by :func:`~hwilib.commands.enumerate`
        :param password: A password/passphrase to use with the device.
            Typically a BIP 39 passphrase, but not always.
            See device specific documentation for further details.
        :param expert: Whether to return additional information intended for experts.
        """
        super().__init__(path, password, expert)
        self.simulator = ":" in path
        if self.simulator:
            self.dev = SpecterSimulator(path)
        else:
            self.dev = SpecterUSBDevice(path)

    def query(self, data: str, timeout: Optional[float] = None) -> str:
        """Send a text-based query to the device and get back the response"""
        res = self.dev.query(data, timeout)
        if res == "error: User cancelled":
            raise ActionCanceledError("User didn't confirm action")
        elif res.startswith("error: Unknown command"):
            raise UnavailableActionError(res[7:])
        elif res.startswith("error: "):
            raise BadArgumentError(res[7:])
        return res

    def get_master_fingerprint(self) -> bytes:
        """
        Get the master public key fingerprint as bytes.

        Retrieves the fingerprint of the master public key of a device.
        Typically implemented by fetching the extended public key at "m/0h"
        and extracting the parent fingerprint from it.

        :return: The fingerprint as bytes
        """
        return bytes.fromhex(self.query("fingerprint", timeout=self.TIMEOUT))

    def get_master_blinding_key(self) -> str:
        """
        Get the master blinding key as WIF string (according to SLIP77 format).

        :return: The master blinding key as WIF string
        """
        return self.query("slip77")

    def get_pubkey_at_path(self, bip32_path: str) -> ExtendedKey:
        """
        Get the public key at the BIP 32 derivation path.

        :param bip32_path: The BIP 32 derivation path
        :return: The extended public key
        """
        # this should be fast
        xpub = self.query("xpub %s" % bip32_path, timeout=self.TIMEOUT)
        hd = ExtendedKey.deserialize(xpub)
        # Specter returns xpub with a prefix
        # for a network currently selected on the device
        hd.version = (
            b"\x04\x88\xb2\x1e" if self.chain == Chain.MAIN else b"\x04\x35\x87\xcf"
        )
        return hd

    def sign_b64psbt(self, psbt: str) -> str:
        # works with both PSBT and PSET
        return self.query("sign %s" % psbt)

    def sign_pset(self, pset: str) -> str:
        return self.sign_b64psbt(pset)

    def sign_tx(self, psbt: PSBT) -> PSBT:
        """
        Sign a partially signed bitcoin transaction (PSBT).

        :param psbt: The PSBT to sign
        :return: The PSBT after being processed by the hardware wallet
        """
        response = self.query("sign %s" % psbt.serialize())
        signed_psbt = PSBT()
        signed_psbt.deserialize(response)
        # adding partial sigs to initial tx
        for i in range(len(psbt.inputs)):
            for k in signed_psbt.inputs[i].partial_sigs:
                psbt.inputs[i].partial_sigs[k] = signed_psbt.inputs[i].partial_sigs[k]
        return psbt

    def sign_message(self, message: Union[str, bytes], bip32_path: str) -> str:
        """
        Sign a message (bitcoin message signing).

        Signs a message using the legacy Bitcoin Core signed message format.
        The message is signed with the key at the given path.

        :param message: The message to be signed. First encoded as bytes if not already.
        :param bip32_path: The BIP 32 derivation for the key to sign the message with.
        :return: The signature
        """
        # convert message to bytes
        msg = message
        if isinstance(message, str):
            msg = message.encode("utf-8")
        # check if ascii - we only support ascii characters display
        try:
            msg.decode("ascii")
            fmt = "ascii"
        except:
            fmt = "base64"
        # check if there is \r or \n in the message
        # in this case we need to encode to base64
        if b"\r" in msg or b"\n" in msg:
            fmt = "base64"
        # convert to base64 if necessary
        if fmt == "base64":
            msg = b2a_base64(msg).strip()
        sig = self.query(f"signmessage {bip32_path} {fmt}:{msg.decode()}")
        return sig

    def display_singlesig_address(
        self,
        bip32_path: str,
        addr_type: AddressType,
    ) -> str:
        """
        Display and return the single sig address of specified type
        at the given derivation path.

        :param bip32_path: The BIP 32 derivation path to get the address for
        :param addr_type: The address type
        :return: The retrieved address also being shown by the device
        """
        if addr_type == AddressType.LEGACY:
            script_type = "pkh"
        elif addr_type == AddressType.SH_WIT:
            script_type = "sh-wpkh"
        elif addr_type == AddressType.WIT:
            script_type = "wpkh"
        else:
            raise BadArgumentError("Invalid address type")
        # prepare a request of the form like
        # `showaddr sh-wsh m/1h/2h/3 descriptor`
        request = f"showaddr {script_type} {bip32_path}"
        address = self.query(request)
        return address

    def display_multisig_address(
        self,
        addr_type: AddressType,
        multisig: MultisigDescriptor,
    ) -> str:
        """
        Display and return the multisig address of specified type given the descriptor.

        :param addr_type: The address type
        :param multisig: A :class:`~hwilib.descriptor.MultisigDescriptor` that describes the multisig to display.
        :return: The retrieved address also being shown by the device
        """
        # prepare a request of the form like
        # `showaddr sh-wsh m/1h/2h/3 descriptor`
        if addr_type == AddressType.LEGACY:
            script_type = "sh"
        elif addr_type == AddressType.SH_WIT:
            script_type = "sh-wsh"
        elif addr_type == AddressType.WIT:
            script_type = "wsh"
        else:
            raise BadArgumentError("Invalid address type")

        script, *_ = multisig.expand(0)
        bip32_path = (
            multisig.pubkeys[0].origin.to_string() + multisig.pubkeys[0].deriv_path
        )
        request = f"showaddr {script_type} {bip32_path} {script.hex()}"
        address = self.query(request)
        return address

    def close(self) -> None:
        # nothing to do for DIY
        pass

    ############ extra functions Specter supports ############

    def get_random(self, num_bytes: int = 32):
        if num_bytes < 0 or num_bytes > 10000:
            raise BadArgumentError("We can only get up to 10k bytes of random data")
        res = self.query("getrandom %d" % num_bytes)
        return bytes.fromhex(res)

    def import_wallet(self, name: str, descriptor: str):
        self.query("addwallet {name} {descriptor}")


def enumerate(password=""):
    """
    Returns a list of detected Specter devices
    with their fingerprints and client's paths
    """
    results = []
    # find ports with micropython's VID
    ports = [
        port.device
        for port in serial.tools.list_ports.comports()
        if is_micropython(port)
    ]
    try:
        # check if there is a simulator on port 8789
        # and we can connect to it
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(("127.0.0.1", 8789))
        s.close()
        ports.append("127.0.0.1:8789")
    except ConnectionRefusedError as e:
        # simulator is probably not running
        pass

    for port in ports:
        # for every port try to get a fingerprint
        try:
            path = port
            data: Dict[str, Any] = {}
            data["type"] = "specter"
            data["model"] = "specter-diy"
            data["path"] = path
            data["needs_pin_sent"] = False
            data["needs_passphrase_sent"] = False
            client = SpecterClient(path, "", False)
            data["fingerprint"] = client.get_master_fingerprint().hex()
            client.close()
            results.append(data)
        except Exception as e:
            print(e)
    return results


############# Helper functions and base classes ##############


def is_micropython(port):
    return "VID:PID=F055:" in port.hwid.upper()


class SpecterBase:
    """Class with common constants and command encoding"""

    EOL = b"\r\n"
    ACK = b"ACK"
    ACK_TIMOUT = 3

    def prepare_cmd(self, data):
        """
        Prepends command with 2*EOL and appends EOL at the end.
        Double EOL in the beginning makes sure all pending data
        will be cleaned up.
        """
        return self.EOL * 2 + data.encode("utf-8") + self.EOL


class SpecterUSBDevice(SpecterBase):
    """
    Base class for USB device.
    Implements a simple query command over serial
    """

    def __init__(self, path):
        self.ser = serial.Serial(baudrate=115200, timeout=30)
        self.ser.port = path

    def read_until(self, eol, timeout=None):
        t0 = time.time()
        res = b""
        while not (eol in res):
            try:
                raw = self.ser.read(1)
                res += raw
            except Exception as e:
                time.sleep(0.01)
            if timeout is not None and time.time() > t0 + timeout:
                self.ser.close()
                raise DeviceBusyError("Timeout")
        return res

    def query(self, data, timeout=None):
        # non blocking
        self.ser.timeout = 0
        self.ser.open()
        self.ser.write(self.prepare_cmd(data))
        # first we should get ACK
        res = self.read_until(self.EOL, self.ACK_TIMOUT)[: -len(self.EOL)]
        # then we should get the data itself
        if res != self.ACK:
            self.ser.close()
            raise DeviceBusyError("Device didn't return ACK")
        res = self.read_until(self.EOL, timeout)[: -len(self.EOL)]
        self.ser.close()
        return res.decode()


class SpecterSimulator(SpecterBase):
    """
    Base class for the simulator.
    Implements a simple query command over tcp/ip socket
    """

    def __init__(self, path):
        arr = path.split(":")
        self.sock_settings = (arr[0], int(arr[1]))

    def read_until(self, s, eol, timeout=None):
        t0 = time.time()
        res = b""
        while not (eol in res):
            try:
                raw = s.recv(1)
                res += raw
            except Exception as e:
                time.sleep(0.01)
            if timeout is not None and time.time() > t0 + timeout:
                s.close()
                raise DeviceBusyError("Timeout")
        return res

    def query(self, data, timeout=None):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(self.sock_settings)
        s.send(self.prepare_cmd(data))
        s.setblocking(False)
        # we will get ACK right away
        res = self.read_until(s, self.EOL, self.ACK_TIMOUT)[: -len(self.EOL)]
        if res != self.ACK:
            raise DeviceBusyError("Device didn't return ACK")
        # fetch with required timeout
        res = self.read_until(s, self.EOL, timeout)[: -len(self.EOL)]
        s.close()
        return res.decode()


###### test for communication ######

if __name__ == "__main__":
    import sys

    devices = enumerate()
    if len(devices) == 0:
        print("No devices found")
        sys.exit()
    inp = 0
    if len(devices) > 1:
        print("Found %d devices." % len(devices))
        for i, dev in py_enumerate(devices):
            print(f"[{i}] {dev['path']} - {dev['fingerprint']}")
        inp = -1
        while True:
            inp = int(input("Enter the device number to use: "))
            if inp >= len(devices) or inp < 0:
                print("Meh... Try again.")
                continue
            break
    dev = SpecterClient(devices[inp]["path"])
    if len(sys.argv) == 1:
        mfp = dev.get_master_fingerprint().hex()
        derivation = "m/84h/0h/0h"
        xpub = dev.get_pubkey_at_path(derivation).to_string()
        print(f"Device fingerprint: {mfp}")
        print(f"Segwit xpub: {xpub}")
        print(f"Full key: [{mfp}{derivation[1:]}]{xpub}")
    else:
        if "-i" not in sys.argv:
            cmd = " ".join(sys.argv[1:])
            print("Running command:", cmd)
            print(dev.query(cmd))
        else:
            cmd = ""
            print("Interactive mode! Enter `quit` to exit.")
            while inp != "quit":
                cmd = input("Enter command to run: ")
                if cmd == "quit":
                    sys.exit(0)
                try:
                    print(dev.query(cmd))
                except Exception as e:
                    print("Error:", e)



import logging
import os
from http.client import HTTPConnection

import click

from .cli_noded import bitcoind, elementsd
from .cli_ext import ext
from .cli_server import server
from .cli_gunicorn import gunicorn

logger = logging.getLogger(__name__)


@click.group()
@click.option("--debug", is_flag=True, help="Show debug information on errors.")
@click.option(
    "--tracerpc/--no-tracerpc",
    default=False,
    help="Will trace all calls to BitcoinCore or ElementsCore if in --debug",
)
@click.option(
    "--tracerequests/--no-tracerequests",
    default=False,
    help="Will trace all calls done via the requests module. Might be quite verbose!",
)
@click.pass_context
def entry_point(config_home, debug=False, tracerpc=False, tracerequests=False):
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    if tracerpc or tracerequests:
        if tracerpc:
            debug = True  # otherwise this won't work
            logging.getLogger("cryptoadvance.specter.rpc").setLevel(logging.DEBUG)
        if tracerequests:
            # from here: https://stackoverflow.com/questions/16337511/log-all-requests-from-the-python-requests-module
            HTTPConnection.debuglevel = 1
            requests_log = logging.getLogger("requests.packages.urllib3")
            requests_log.setLevel(logging.DEBUG)
            requests_log.propagate = True
    else:
        logging.getLogger("cryptoadvance.specter.rpc").setLevel(logging.INFO)

    if debug:
        # No need for timestamps while developing
        formatter = logging.Formatter("[%(levelname)7s] in %(module)15s: %(message)s")
        logging.getLogger("cryptoadvance").setLevel(logging.DEBUG)
        # but not that chatty connectionpool
        logging.getLogger("urllib3.connectionpool").setLevel(logging.INFO)
    else:
        formatter = logging.Formatter(
            # Too early to format that via the flask-config, so let's copy it from there:
            os.getenv(
                "SPECTER_LOGFORMAT",
                "[%(asctime)s] %(levelname)s in %(module)s: %(message)s",
            )
        )
        logging.getLogger("cryptoadvance").setLevel(logging.INFO)
    ch.setFormatter(formatter)
    logging.getLogger().handlers = []
    logging.getLogger().addHandler(ch)


entry_point.add_command(server)
entry_point.add_command(gunicorn)
entry_point.add_command(ext)
entry_point.add_command(bitcoind)
entry_point.add_command(elementsd)
